#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2018 Centrify Corporation. All rights reserved.
#
# RHEL-specific machine mapper script to import coolkey to firefox
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::General qw(:debug RunCommand RunCmd CreateDir CreateFile ReadFile);
use CentrifyDC::GP::Mapper;
use CentrifyDC::GP::RegHelper;
use POSIX qw(setuid setgid);

# Global values
my $REGKEY_ROOT = 'software/policies/centrify/unixsettings/security/importnssdb';
my $SC_REGKEY = 'software/policies/centrify/unixsettings/security/rhelsmartcardenabled';

my $ARCHTYPE = `uname -m`;
chomp($ARCHTYPE);
if ($ARCHTYPE eq 'x86_64')
{
    $ARCHTYPE = 'lib64';
}
else
{
    $ARCHTYPE = 'lib';
}

# the name of the module to install/uninstall
my $PKCS11_MODULE_NAME = 'Centrify PKCS #11 Module';
my $PKCS11_MODULE_PATH = '/usr/' . $ARCHTYPE . '/pkcs11/libcentrifypkcs11.so';

# the user's home directory
my $homedir;
# array of all applications that want to importnssdb
my $settings;
# user object from getpwnam
my @user;


# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# Set global values that are used to install and uninstall coolkey
sub SetGlobals($)
{
    my ($username) = @_;
    @user = getpwnam($username);
    $homedir = $user[7];
    defined($homedir) or FATAL_OUT("Cannot get home dir of user [$username]");
}

sub RemoveLegacyModule($)
{
    my ($dbdir) = @_;
    my ($ret, $rawlist) = RunCmd(["modutil", "-rawlist", "-dbdir", $dbdir]);
    $ret = $rawlist;
    if ($ret =~ m/.*Centrify_Coolkey.*/)
    {
        $rawlist =~ m/.*Centrify_Coolkey/g;
        my $ck_position = pos($rawlist) - length("Centrify_Coolkey");
        my $ck_name = substr $rawlist, $ck_position;

        # get rid of white spaces at the end
        $ck_name =~ m/^\S+\s/g;
        $ck_position = pos($ck_name) - 1;
        $ck_name = substr $ck_name, 0, $ck_position;
        DEBUG_OUT("Removing old coolkey module: [$ck_name]");
        `modutil -delete $ck_name -dbdir $dbdir -force`; 
    } 
}

sub UninstallCoolkey($)
{
    DEBUG_OUT("Uninstall coolkey");

    my ($username) = @_;
    my $pid;
    defined ($pid = fork) or die "Cannot fork $!";
    if ($pid < 0)
    {
        die "Fork failed $!";
    }
    if ($pid == 0)
    {
        # child process
        if (@user)
        {
            DEBUG_OUT("Setting to gid: [$user[3]]");
            DEBUG_OUT("Setting to uid: [$user[2]]");
            setgid $user[3] or die "Cannot switch group for $username $!";
            setuid $user[2] or die "Cannot switch id to $username $!";
        }
        else
        {
            die "Invalid user ID($username)";
        }
        foreach my $setting (@$settings)
        {
            DEBUG_OUT("The setting is currently: [$setting]");
            # cut off the first character which should be the tilde
            my $firstChar = substr($setting, 0, 1);
            if ($firstChar eq '~')
            {
                $setting = substr($setting, 1);
            }
            else
            {
                DEBUG_OUT("The first character of this user setting is [$firstChar] from [$setting] which is not a [~],
                    continuing to the next setting");
                next;
            }
            my $new_app_location = $homedir . $setting;
            DEBUG_OUT("Uninstalling coolkey lib from [$new_app_location]");
            my $new_app_profiles_ini = $new_app_location . "/profiles.ini"; 
            if (CreateDir($new_app_location) && CreateFile($new_app_profiles_ini))
            {
                my $module;
                my $str = ReadFile($new_app_profiles_ini);
                if (! defined($str))
                {
                    DEBUG_OUT("Cannot read $new_app_profiles_ini");
                    next;
                }
                
                my @lines = split(/\n/, $str);
                
                foreach my $line (@lines)
                {
                    chomp $line;
                    if ($line =~ m/^Path=.*/)
                    {
                        $line =~ s/^Path=//;
                        RemoveLegacyModule("$new_app_location/$line");

                        $module = `modutil -rawlist -dbdir $new_app_location/$line`;
                        if ($module =~ m/.*$PKCS11_MODULE_NAME.*/)
                        {
                            DEBUG_OUT("Removing $PKCS11_MODULE_NAME: [$PKCS11_MODULE_PATH]");
                            RunCmd(["modutil", "-delete", $PKCS11_MODULE_NAME, "-dbdir", "$new_app_location/$line", "-force"]);
                        }
                    }
                }
            }
        }
        exit(1);
    }
    else
    {
        # parent process
        if (waitpid($pid, 0) < 0)
        {
            die "waitpid failed for pid $pid";
        }
    }
}

sub InstallCoolkey($)
{
    DEBUG_OUT("Install coolkey");
    
    # load coolkey module to all firefox profiles
    my ($username) = @_;
    my $pid;
    DEBUG_OUT("user: [$username] has the homedir: [$homedir]");
    defined ($pid = fork) or die "Cannot fork $!";
    if ($pid < 0)
    {
        die "Fork failed $!";
    }
    if ($pid == 0)
    {
        # child process
        if (@user)
        {
            DEBUG_OUT("Setting to gid: [$user[3]]");
            DEBUG_OUT("Setting to uid: [$user[2]]");
            setgid $user[3] or die "Cannot switch group for $username $!";
            setuid $user[2] or die "Cannot switch id to $username $!";
        }
        else
        {
            die "Invalid user ID($username)";
        }
        foreach my $setting (@$settings)
        {
            DEBUG_OUT("The setting is currently: [$setting]");
            # cut off the first character which should be the tilde
            my $firstChar = substr($setting, 0, 1);
            if ($firstChar eq '~')
            {
                $setting = substr($setting, 1);
            }
            else
            {
                DEBUG_OUT("The first character of this user setting is [$firstChar] from [$setting] which is not a [~],
                    continuing to the next setting");
                next;
            }
            my $new_app_location = $homedir . $setting;
            DEBUG_OUT("Installing [$PKCS11_MODULE_NAME] to [$new_app_location]");
            my $new_app_profiles_ini = $new_app_location . "/profiles.ini"; 
            if (CreateDir($new_app_location) && CreateFile($new_app_profiles_ini))
            {
                my $str = ReadFile($new_app_profiles_ini);
                if (! defined($str))
                {
                    DEBUG_OUT("Cannot read $new_app_profiles_ini");
                    next;
                }
                
                my @lines = split(/\n/, $str);
                
                foreach my $line (@lines)
                {
                    if ($line =~ m/^Path=.*/)
                    {
                        $line =~ s/^Path=//;
                        my $module = `modutil -rawlist -dbdir $new_app_location/$line`;

                        # remove the old centrify_coolkey module if any
                        RemoveLegacyModule("$new_app_location/$line");

                        if ($module =~ m/.*$PKCS11_MODULE_NAME.*/ && $module =~ m/.*$PKCS11_MODULE_PATH.*/)
                        {
                            DEBUG_OUT("Not importing $PKCS11_MODULE_NAME because we found the same module in [$new_app_location/$line]");
                        }
                        else
                        {
                            DEBUG_OUT("Importing $PKCS11_MODULE_NAME: [$PKCS11_MODULE_PATH] lib: [$ARCHTYPE] to: [$new_app_location/$line]");
                            RunCommand("env LD_LIBRARY_PATH= modutil -add '$PKCS11_MODULE_NAME' -libfile $PKCS11_MODULE_PATH -dbdir $new_app_location/$line -force");
                        }
                    }
                }
            }
        }
        exit(1);
    }
    else
    {
        # parent process
        if (waitpid($pid, 0) < 0)
        {
            die "waitpid failed for pid $pid";
        }
    }
}

sub GetSetting($$)
{
    my ($user, $action) = @_;
    # only allow the coolkey module to be installed or uninstalled based on if rhel smartcard is enabled
    my $screg = CentrifyDC::GP::RegHelper->new($action, 'machine', $SC_REGKEY, 'RHELSmartCardEnabled', undef);
    $screg->load();
    my $scsetting = $screg->get('current');
    CentrifyDC::GP::Registry::Load($user);
    my $reg = CentrifyDC::GP::RegHelper->new($action, 'user', $REGKEY_ROOT, undef, undef, 1);
    $reg->load();
    # FIXME: removing individual apps in the GP will not cause those apps to
    # actually uninstall the coolkey module when this script is run
    $settings = $reg->get('current');
    if ($action eq 'unmap')
    {
        SetGlobals($user);
        # Just a workaround... see above FIXME
        $settings = $reg->get('previous');
        UninstallCoolkey($user);
    }
    elsif ($scsetting eq "YES" && $action eq 'map')
    {
        SetGlobals($user);
        if (defined($settings) && scalar(@$settings) > 0)
        {
            InstallCoolkey($user);
        }
        else
        {
            # Just a workaround... see above FIXME
            $settings = $reg->get('previous');
            UninstallCoolkey($user);
        }
    }
    else
    {
        DEBUG_OUT("Smart card for RHEL was not enabled, not installing coolkey libraries");
    }
    return 1;
}

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $err = 0;
my $args = CentrifyDC::GP::Args->new('user');
CentrifyDC::GP::Registry::Load(undef);

if ($args->isMap())
{
    GetSetting($args->user(), $args->action) or FATAL_OUT("Can't get registry settings");
}
else
{
    GetSetting($args->user(), $args->action) or FATAL_OUT("Can't get registry settings");
}

$err and FATAL_OUT();
