##############################################################################
#
# Copyright (C) 2004-2018 Centrify Corporation. All rights reserved.
#
# Centrify DirectControl mapper script certificate utility module.
#
##############################################################################

use strict;

package CentrifyDC::GP::CertUtil;
my $VERSION = "1.0";
require 5.000;

use vars qw(@ISA @EXPORT_OK);

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(GetFingerprint GetIssuingCA GetCADNSHostName IsValid);

use CentrifyDC::GP::General qw(:debug RunCmd EscapeSingleQuote);
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);

my $openssl = "$CIMS_INSTALL/bin/openssl";
my $ldapsearch = "$CIMS_INSTALL/bin/ldapsearch";
my $adcert = "$CIMS_INSTALL/sbin/adcert";

sub GetFingerprint($;$);
sub GetIssuingCA($);
sub GetCADNSHostName($);
sub IsValid($$);


#
# Get fingerprint from a certificate file which is in specific format. Supported
# format are PEM, DER and NET.
#
#   $_[0]:  cert file
#   $_[1]:  cert file format
#
#   return: certificate fingerprint
#           undef   - failed
#
sub GetFingerprint($;$)                                                                                                                                                                                                                                                                                                   
{
    my ($file, $certForm) = @_; 

    if ($certForm ne 'PEM' && $certForm ne 'DER' && $certForm ne 'NET')
    {
        $certForm = 'DER';
    }

    TRACE_OUT("Get fingerprint from [$file]");
    #   
    # get fingerprint from certificate file.
    # output format:
    # SHA1 Fingerprint=XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX
    # to extract fingerprint, remove heading and :
    #   
    # Do not change the way the cache file name is determined (with
    # openssl x509 -fingerprint -sha1) without a corresponding change
    # to lib/darwin/pkinit.cpp.
    #
    my $cmd = [$openssl, "x509", "-fingerprint", "-sha1", "-noout", "-inform", $certForm, "-in", $file];
    my ($rc, $output) = RunCmd($cmd);
    if (! defined($rc) or $rc ne '0')
    {   
        ERROR_OUT("Cannot get fingerprint from $file");
        return undef;
    }   
    my $fingerprint = $output;
    $fingerprint =~ s/.*Fingerprint=//;
    $fingerprint =~ s/://g;
    chomp($fingerprint);
    
    if (defined($fingerprint))
    {   
        return $fingerprint;
    }   

    return undef;
}

sub GetCADNSHostName($)
{
    my ($caName) = @_;
    my $val;
    my $dcdom;
    
    my ($ret, $out) = RunCmd([$ldapsearch, "-m", "-LLL", "-s", "base", "-b", "", "-r", "configurationNamingContext"]);
    foreach $val (split /\n/, $out)
    {
        if ($val =~ /^configurationNamingContext: (.*)$/)
        {
            $dcdom = $1;
            last;
        }
    }
    
    my $cmd = [$ldapsearch, "-r", "-LLL", "-m", "-b", "CN=Enrollment Services,CN=Public Key Services,CN=Services,${dcdom}", "(name=${caName})", "dNSHostName"];
    
    ($ret, $out) = RunCmd($cmd);
    
    my @ldap = split /\n/, $out;
    foreach $val (@ldap)
    {
        if ($val =~ /^dNSHostName: (.*)/)
        {
            return $1;
        }
    }
    
    return undef;
}

sub GetIssuingCA($)
{
    my ($templateName) = @_;
    
    my $val;
    my $dcdom;
    
    my ($ret, $out) = RunCmd([$ldapsearch, "-m", "-LLL", "-s", "base", "-b", "", "-r", "configurationNamingContext"]);
    foreach $val (split /\n/, $out)
    {
        if ($val =~ /^configurationNamingContext: (.*)$/)
        {
            $dcdom = $1;
            last;
        }
    }
    
    
    my $cmd = [$ldapsearch, "-r", "-LLL", "-m", "-b", "CN=Enrollment Services,CN=Public Key Services,CN=Services,${dcdom}", "(certificateTemplates=${templateName})", "name"];
    ($ret, $out) = RunCmd($cmd);
    
    my @ldap = split /\n/, $out;
    my @ca = ();
    foreach $val (@ldap)
    {
        if ($val =~ /^name: (.*)/)
        {
            push(@ca, $1);
        }
    }
    
    return \@ca;
}


sub IsValid($$)
{
    my ($certPath, $templateName) = @_;
    my $cmd = [$adcert, "-x", "-c", $certPath, "-t", $templateName];
    
    my ($rc, $out) = RunCmd($cmd);
    if ($rc == 156)
    {
        DEBUG_OUT("Certificate $certPath is not valid: $out");
        return 0;
    }
    
    return 1;
}


1;
