##############################################################################
#
# Copyright (C) 2004-2018 Centrify Corporation. All rights reserved.
#
# Centrify DirectControl mapper script general purpose module.
#
##############################################################################

use strict;

package CentrifyDC::GP::General;
my $VERSION = '1.0';
require 5.000;

use vars qw(@ISA @EXPORT_OK %EXPORT_TAGS $GP_DEBUG $GP_TRACE $GP_LOG_PRINT_TO_STDOUT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(FATAL_OUT ERROR_OUT WARN_OUT INFO_OUT DEBUG_OUT TRACE_OUT IsDebugOn IsTraceOn
                RunCommand RunCommandWithTimeout RunCmd ChangeOwner CreateDir CreateFile ReadFile CopyFile WriteFile IsFileEmpty GetFullPath GetTempDirPath CreateTempFile TraverseSymLink
                IsEqual IsEmpty  AddElementsIntoArray RemoveElementsFromArray DiffTwoArray EscapeSingleQuote IsFipsSupported IsFipsEnabled);
%EXPORT_TAGS = (
    'debug'     => [qw(FATAL_OUT ERROR_OUT WARN_OUT INFO_OUT DEBUG_OUT TRACE_OUT IsDebugOn IsTraceOn)],
    'system'    => [qw(RunCommand RunCommandWithTimeout RunCmd ChangeOwner CreateDir CreateFile ReadFile CopyFile WriteFile IsFileEmpty GetFullPath GetTempDirPath CreateTempFile TraverseSymLink)],
    'general'   => [qw(IsEqual IsEmpty AddElementsIntoArray RemoveElementsFromArray DiffTwoArray EscapeSingleQuote IsFipsSupported IsFipsEnabled)]);

use POSIX qw(:sys_wait_h :signal_h);

use IPC::Open3;
use File::Spec;
use Symbol qw(gensym);

use Fcntl qw(LOCK_EX LOCK_NB LOCK_UN O_RDONLY O_RDWR O_CREAT O_EXCL O_NONBLOCK);
use File::Basename qw(basename dirname);
use File::Path qw(mkpath);
use File::Copy qw(move copy);
use File::stat qw(lstat);

use CentrifyDC::Logger;
use CentrifyDC::GP::Lock;

#
# other than the system wide log level in centrifydc.conf, we can also set
# log settings for all gp mappers or an individual mapper.
#
# to set gp mappers log level to DEBUG, and mac_mapper_network.pl to TRACE:
#
#   log: INFO
#   log.gp.mappers: DEBUG
#   log.gp.mappers.mac_mapper_network.pl: TRACE
#
# if log level is TRACE, then the output of system command will also be logged
# (need to use RunCommand)
#
# sometimes developer may want to print log message to stdout:
#
#   gp.mappers.print_log_to_stdout: 1
#
# or
#
#   gp.mappers.print_log_to_stdout.mac_mapper_network.pl: 1
#
$GP_DEBUG = 0;              # debug level
$GP_TRACE = 0;              # trace level (more information)
$GP_LOG_PRINT_TO_STDOUT = 0; # print all log entries to stdout. for debug

my $KILL_TIMEOUT = 2;
my $MAX_OUTPUT_LINES = 1000;


my $TEMP_DIR_NON_ROOT_WRITABLE_PATH = '/tmp';
my $TEMP_DIR_NON_ROOT_READABLE_PATH = '/var/centrify/tmp';
my $TEMP_DIR_NON_ROOT_READABLE_PERM = 0755;
my $TEMP_DIR_NON_ROOT_READABLE_UID = 0;
my $TEMP_DIR_NON_ROOT_READABLE_GID = 0;

# since Perl 5.8.8, chdir/chmod/chown also supports file handler. refer to:
# http://perldoc.perl.org/5.8.8/perldelta.html
my $IS_CHXXX_SUPPORT_FH = ($] ge '5.008008');

#
# Define if the directory path should be skipped for path fixing
#
#   hash ref key:       full directory path
#   hash ref value:     can be skipped (1) or not (0)
#
#   e.g. '/var' => '1',
#
my $fix_path_exception_list = {
    '/var' => '1',
};

my $TEMP_FILE_NAME_TRY_COUNT = 10;

my $MAX_FILE_SIZE = 256000000;  # max file size for ReadFile/WriteFile

my $gp_logger;

sub __InitLogger();

# debug
sub FATAL_OUT;  # fatal error occured. exit 1. tell user how to enable debug log if not enabled
sub ERROR_OUT;  # log error message
sub WARN_OUT;   # log warning message
sub INFO_OUT;   # log info message
sub DEBUG_OUT;  # log debug message
sub TRACE_OUT;  # log trace message
sub IsDebugOn();
sub IsTraceOn();
sub _Log;

# system
sub _GetFileOptions($);
sub _RestoreCon($);
sub _TimedWait($$);
sub RunCommandWithTimeout($$);
sub RunCommand($;$);
sub RunCmd($;$);
sub _FindCmdInPath($);
sub _RunCmdWithSystem($);
sub _RunCmdWithPipe($);
sub _RunCmdWithOpen3($$);
sub ChangeOwner($$);    # change file's owner to specified user
sub CreateDir($;$);     # create dir and set owner
sub CreateFile($;$);    # create empty file and set owner
sub ReadFile($);        # read file into a string
sub CopyFile($$);       # copy file
sub WriteFile($$;$);    # write string into file
sub IsFileEmpty($);     # check if file is empty
sub GetFullPath($);     # return full path of a given file
sub GetTempDirPath($);  # return temp dir path
sub CreateTempFile($);  # create a temp file and return file handler/name
sub TraverseSymLink($); # traverse symlink and return the actual file

# general
sub IsEqual;    # compare two variable/hash/array/scalar
sub _IsEqual;
sub IsEmpty($);
sub AddElementsIntoArray($$);
sub RemoveElementsFromArray($$);
sub DiffTwoArray($$);
sub EscapeSingleQuote($);
sub IsFipsSupported;    # Check if FIPS mode is supported
sub IsFipsEnabled;      # Check if FIPS mode is enabled



#
# get log setting from centrifydc.conf and init logger.
#
sub __InitLogger()
{
    my $program_name = basename $0;

    $gp_logger = CentrifyDC::Logger->new('');

    my @log_entries = ('log', 'log.gp.mappers', "log.gp.mappers.${program_name}");
    foreach my $log_entry (@log_entries)
    {
        my $property = $CentrifyDC::Config::properties{$log_entry};
        if (defined($property))
        {
            $gp_logger->level($property);
            if ($property eq 'DEBUG')
            {
                $GP_DEBUG = 1;
                $GP_TRACE = 0;
            }
            elsif ($property eq 'TRACE')
            {
                $GP_DEBUG = 1;
                $GP_TRACE = 1;
            }
            else
            {
                $GP_DEBUG = 0;
                $GP_TRACE = 0;
            }
        }
    }

    my @redirect_entries = ('gp.mappers.print_log_to_stdout', "gp.mappers.print_log_to_stdout.${program_name}");
    foreach my $redirect_entry (@redirect_entries)
    {
        my $property = $CentrifyDC::Config::properties{$redirect_entry};
        if (defined($property))
        {
            if ($property eq '0')
            {
                $GP_LOG_PRINT_TO_STDOUT = 0;
            }
            else
            {
                $GP_LOG_PRINT_TO_STDOUT = 1;
            }
        }
    }
}



# >>> DEBUG >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# log fatal error message and exit 1. if debug log is not enabled, tell user how to
# enable it.
#
#   @_:  message (optional)
#
#   exit:   1
#
sub FATAL_OUT
{
    my $program_name = basename $0;

    my $loglevel = 'info';

    my $first = shift;
    if (defined($first))
    {
        $first = ">>> " . $first;
        unshift(@_, $first);
        _Log($loglevel, @_);
    }
    else
    {
        _Log($loglevel, ">>> A problem occured while running $0");
    }

    if (! $GP_DEBUG)
    {
        _Log($loglevel, 'Please enable debug log by adding the following line into /etc/centrifydc/centrifydc.conf:');
        _Log($loglevel, "log.gp.mappers.${program_name}\: DEBUG");
        _Log($loglevel, 'Use TRACE instead of DEBUG will give you the most detailed information.');
    }

    exit(1);
}

#
# log error message
#
# filename, line number, and parent subroutine will be added in front of the
# actual error message
#
#   @_:  message
#
sub ERROR_OUT
{
    my $first = shift;

    my $filename = (caller(0))[1];
    my $line = (caller(0))[2];
    my $subroutine = defined(caller(1)) ? (caller(1))[3] : 'main';

    $first = ">>> $filename : $line : $subroutine : $first";
    unshift(@_, $first);

    _Log('info', @_);
}

sub WARN_OUT
{
    _Log('info', @_);
}

sub INFO_OUT
{
    _Log('info', @_);
}

sub DEBUG_OUT
{
    $GP_DEBUG and _Log('debug', @_);
}

sub TRACE_OUT
{
    $GP_TRACE and _Log('trace', @_);
}

sub IsDebugOn()
{
    return $GP_DEBUG;
}

sub IsTraceOn()
{
    return $GP_TRACE;
}

#
# log message and print to stdout if necessary
#
#   $_[0]:  log level
#   @_:  log message
#
sub _Log
{
    defined($_[1]) or return;

    my $loglevel = shift;
    my $format = shift;

    # For outputing the quotes in loginfo,we must use "\" to escape the quotes.
    my $msg = sprintf($format, @_);
    $msg =~ s/\"/\\"/g;
    $gp_logger->log($loglevel, $msg);

    $GP_LOG_PRINT_TO_STDOUT or return;
    print STDOUT "$msg\n";
}

# <<< DEBUG <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<



# >>> SYSTEM >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# Wait for the specified number of seconds for a child process to finish.
#
#   $_[0]:  pid
#   $_[1]:  timeout (seconds)
#
sub _TimedWait($$)
{
    my ($pid, $timeout) = @_;

    my ($ret, $error);

    eval
    {
        local $SIG{ALRM} = sub { die "alarm\n" };
        alarm $timeout;
        my $rc = waitpid($pid, 0);
        $ret = $? >> 8;
        alarm 0;
    };

    $error = $@;

    if ($error)
    {
        if ($error ne "alarm\n")   # propagate unexpected errors
        {
            ERROR_OUT("process $pid died unexpectedly: $error");
        }
    }

   return ($ret, $error);
}

#
# run command with specified timeout (deprecated)
#
# NOTICE: this command is deprecated. Please use the more secure version
#         RunCmd.
#
# if command doesn't finish in specified time, send SIGTERM and wait for 2
# seconds; if still doesn't finish, send SIGKILL and wait for 2 seconds;
# if still doesn't finish, fail.
#
# in scalar context, return command return value.
# in list context, return command return value and command output.
# If lock seed is specified, a lockfile will be created to ensure
# no other process is running the same command. lockfile name is
# generated from the seed
#
#   $_[0]:  command
#   $_[1]:  timeout (seconds)
#
#   return: 
#       in scalar context:
#           number  - return value of the command
#           undef   - failed
#       in list context:
#       $1: 
#           number  - return value of the command
#           undef   - failed
#       $2:
#           string  - command output
#
sub RunCommandWithTimeout($$)
{
    my ($cmd, $timeout) = @_;

    defined($cmd) or return undef;
    defined($timeout) or return undef;

    DEBUG_OUT("Run command [%s] with timeout [%s] seconds", $cmd, $timeout);

    my $pid = open3(gensym, \*PH, \*PH, $cmd);

    my $iszombie = 0;
    my ($ret, $error) = _TimedWait($pid, $timeout);

    if ($error)
    {
        DEBUG_OUT("Command [%s] taking too long, sending SIGTERM", $cmd);
        kill(SIGTERM, $pid);
    
        (undef, $error) = _TimedWait($pid, $KILL_TIMEOUT);
    }

    if ($error)
    {
        DEBUG_OUT("Command [%s] is not responding to SIGTERM, sending SIGKILL", $cmd);
        kill(SIGKILL, $pid);
    
        (undef, $error) = _TimedWait($pid, $KILL_TIMEOUT);
    }

    if ($error)
    {
        ERROR_OUT("Command [%s] is not responding to SIGKILL", $cmd);
        $iszombie = 1;
    }

    my @output = <PH>;

    my $openfail = 0;
    foreach my $line (@output)
    {
        if ($line =~ m/^open3:/g)
        {
            ERROR_OUT("Fail to run command using open3: %s", $line);
            $openfail = 1;
            last;
        }
    }

    if (! $openfail)
    {
        if (! defined($ret))
        {
            if ($iszombie)
            {
                ERROR_OUT("Command [%s] takes too long to run and cannot be killed", $cmd);
            }
            else
            {
                ERROR_OUT("Command [%s] takes too long to run and is killed", $cmd);
            }
        }
    }

    # only log output when loglevel is TRACE
    if ($GP_TRACE)
    {
        if (! IsEmpty(\@output))
        {
            my $i = 1;
            TRACE_OUT(" command return: $ret");
            TRACE_OUT(" command output:");
            foreach my $line (@output)
            {
                my $str = $line;
                chomp $str;
                TRACE_OUT("%s", " | $str");
                $i++;
                if ($i > $MAX_OUTPUT_LINES)
                {
                    TRACE_OUT("  ... command output too long. truncate.");
                    last;
                }
            }
        }
    }

    wantarray() ? (return ($ret, join('', @output))) : (return $ret);
}

#
# run command (deprecated)
#
# in scalar context, return command return value.
# in list context, return command return value and command output.
# If lock seed is specified, a lockfile will be created to ensure
# no other process is running the same command. lockfile name is
# generated from the seed
#
# NOTICE: this command is deprecated. Please use the more secure version
#         RunCmd.
#
#   $_[0]:  command
#   $_[1]:  lock seed (optional)
#
#   return: 
#       in scalar context:
#           number  - return value of the command
#           undef   - failed
#       in list context:
#       $1: 
#           number  - return value of the command
#           undef   - failed
#       $2:
#           string  - command output
#
#   NOTE:   If your command is constructed using variables and if these 
#           variables might contain special character, please escape 
#           them like this
#           e.g.: $cmd = "adquery user '" . EscapeSingleQuote($user) . "'"
#           
sub RunCommand($;$)
{
    my ($cmd, $lockfile) = @_;

    my $ret;

    TRACE_OUT("Notice: RunCommand is deprecated. Please use RunCmd instead.");
    DEBUG_OUT("run command: [%s]", $cmd);
    if (defined($lockfile))
    {
        TRACE_OUT(" lockfile: [%s]", $lockfile);
        my $lock = CentrifyDC::GP::Lock->new($lockfile, 1);
        if (! defined($lock))
        {
            ERROR_OUT("Cannot obtain lock");
            return undef;
        }
    }

    my $rc;

    # include both STDOUT and STDERR
    $rc = open(OUTPUT, "$cmd 2>&1 |");
    if (! $rc)
    {
        ERROR_OUT("Cannot open pipe for command [%s]: %s", $cmd, $!);
        return undef;
    }

    my @output = <OUTPUT>;

    $rc = close(OUTPUT);
    #
    # need to grab $? right after closing the pipe, because logger module
    # will use system command
    #
    $ret = $? >> 8;
    if ($? & 127)
    {
        # command terminated by signal
        ERROR_OUT("Command [%s] terminated by signal %s", $cmd, $? & 127);
        $ret = undef;
    }
    else
    {
        TRACE_OUT(" return: [$ret]");
    }
    if (! $rc)
    {
        #
        # If the file handle came from a piped open, close will additionally
        # return false if one of the other system calls involved fails,
        # or if the program exits with non-zero status.
        # If the only problem was that the program exited non-zero,
        # $! will be set to 0 .
        #
        if ($!)
        {
            WARN_OUT("Cannot close pipe");
        }
    }


    # only log output when loglevel is TRACE
    if ($GP_TRACE)
    {
        if (! IsEmpty(\@output))
        {
            TRACE_OUT(" command output:");
            foreach my $line (@output)
            {
                my $str = $line;
                chomp $str;
                TRACE_OUT("%s", " | $str");
            }
        }
    }

    wantarray() ? (return ($ret, join("", @output))) : (return $ret);
}

#
# run command
#
# In scalar context, return command return value. Will use "system" to run
# the command unless gp.mappers.force.capture.command.output is true.
#
# In list context, return command return value and command output.
#
# A lockfile can be created to ensure no other process is running the same
# command.
#
#   $_[0]:  array reference of command and arguments, or string of the whole command
#   $_[1]:  hash reference of options (optional)
#           supported options are:
#               lockfile => string, lock file name
#               timeout =>  number, timeout in seconds to terminate command
#
#   return: 
#       in scalar context:
#           number  - return value of the command
#           undef   - failed
#       in list context:
#       $1: 
#           number  - return value of the command
#           undef   - failed
#       $2:
#           string  - command output
#
sub RunCmd($;$)
{
    my ($args, $options) = @_;

    defined($args) or return undef;

    my $is_cmd_array;

    # make a copy of the command array
    my @cmd_args;
    my $cmd_str;

    if (ref($args) eq 'ARRAY')
    {
        $is_cmd_array = 1;
        @cmd_args = @$args;
    }
    elsif (ref($args) eq '')
    {
        $is_cmd_array = 0;
        $cmd_str = $args;
        @cmd_args = ($cmd_str);
    }
    else
    {
        ERROR_OUT("Fail to RunCmd. Args must be an array reference or a string");
        return undef;
    }

    if ($is_cmd_array && scalar($args) < 1)
    {
        ERROR_OUT("Fail to RunCmd. Args must contain at least one item");
        return undef;
    }

    if (defined($options) && ref($options) ne 'HASH')
    {
        ERROR_OUT("Fail to RunCmd. Options must be a hash reference");
        return undef;
    }

    # do we need command output?
    my $want_output = wantarray() ? 1 : 0;
    my $force_output = ($CentrifyDC::Config::properties{'gp.mappers.force.capture.command.output'} eq 'true') ? 1 : 0;

    if ($is_cmd_array)
    {
        # check if command itslef is absolute path. if not, search the command in
        # $PATH, and log it. if cannot find the command, return error.
        my $cmd_file = shift(@cmd_args);
        if ($cmd_file !~ m|^/|)
        {
            DEBUG_OUT("[$cmd_file] is not absolute path. Try to find command in PATH");
            $cmd_file = _FindCmdInPath($cmd_file);
            if (! defined($cmd_file))
            {
                ERROR_OUT("Cannot find $cmd_file in PATH");
                return;
            }
        }

        # generate command string
        $cmd_str = $cmd_file;
        # single quote the command if contains space
        if ($cmd_str =~ m/ /)
        {
            $cmd_str = "'$cmd_str'";
        }
        foreach my $arg (@cmd_args)
        {
            my $arg_str = _ProcessCmdArg($arg);
            if (! defined($arg_str))
            {
                ERROR_OUT("Cannot process command arg [$arg]");
                return;
            }
            $cmd_str .= " $arg_str";
        }

        # put the updated command file back into args
        unshift(@cmd_args, $cmd_file);
    }

    # create optional lock
    my $lock;
    if (defined($options->{lockfile}))
    {
        my $lockfile = $options->{lockfile};
        $lock = CentrifyDC::GP::Lock->new($lockfile, 1);
        if (! defined($lock))
        {
            ERROR_OUT("Cannot obtain lock");
            return;
        }
    }

    DEBUG_OUT("run command: [%s]", $cmd_str);
    # call different function to run command, depending on the requirement

    my $timeout;
    if (exists($options->{timeout}))
    {
        $timeout = $options->{timeout};
    }

    my $output;
    my $ret;

    if ($timeout)
    {
        DEBUG_OUT("Run command using open3");
        # use open3 to run command that require timeout
        ($ret, $output) = _RunCmdWithOpen3($cmd_str, $timeout);
    }
    else
    {
        if ($want_output || $force_output)
        {
            DEBUG_OUT("Run command using pipe");
            ($ret, $output) = _RunCmdWithPipe($cmd_str);
        }
        else
        {
            # try use system to run command as its safer.
            DEBUG_OUT("Run command using system");
            $ret = _RunCmdWithSystem(\@cmd_args);
        }
    }

    wantarray() ? (return ($ret, $output)) : (return $ret);
}

#
# Find full path of a command from PATH
#
#   $_[0]:  command
#
#   return: string  - full path of command
#           undef   - failed
#
sub _FindCmdInPath($)
{
    my $file = $_[0];

    defined($file) or return undef;

    if ($file =~ m|^/|)
    {
        DEBUG_OUT("$file is already absolute path");
        return $file;
    }

    my @path = split(/:/, $ENV{PATH});
    my $full_path;
    my $found = 0;
    foreach my $dir (@path)
    {
        $dir =~ s|/*$||g;
        $dir .= '/';
        $full_path = $dir . $file;
        if (-f $full_path && -x $full_path)
        {
            $found = 1;
            last;
        }
    }

    if ($found)
    {
        DEBUG_OUT("Found command: $full_path");
        return $full_path;
    }
    else
    {
        ERROR_OUT("Cannot find $file in PATH");
        return undef;
    }
}

#
# Process command arg. escape single quote and quote the arg.
#
#   $_[0]:  arg
#
#   return: string  - processed arg
#           undef   - failed
#
sub _ProcessCmdArg($)
{
    my $arg = $_[0];

    defined($arg) or return undef;

    # escape single quote
    if ($arg =~ m/'/)
    {
        $arg =~ s/'/'\\''/g;
    }

    $arg = "'$arg'";

    return $arg;
}

#
# Run command using "system". it won't get command output.
#
#   $_[0]:  array reference of args
#
#   return: number  - return value of the command
#           undef   - failed
#
sub _RunCmdWithSystem($)
{
    my $args = $_[0];

    defined($args) or return undef;

    if (ref($args) ne 'ARRAY')
    {
        ERROR_OUT("args must be an array reference");
        return undef;
    }

    system(@$args);

    my $rc = $?;
    if ($rc == -1)
    {
        ERROR_OUT("Failed to execute: $!");
        return undef;
    }
    elsif ($rc & 127)
    {
        ERROR_OUT("Command terminated by signal %s", $rc & 127);
        return undef;
    }
    else
    {
        $rc = $rc >> 8;
    }

    return $rc;
}

#
# Run command using pipe to get command output (both STDOUT and STDERR).
#
#   $_[0]:  command string
#
#   return: 
#       in scalar context:
#           number  - return value of the command
#           undef   - failed
#       in list context:
#       $1: 
#           number  - return value of the command
#           undef   - failed
#       $2:
#           string  - command output
#
sub _RunCmdWithPipe($)
{
    my $cmd = $_[0];

    defined($cmd) or return undef;

    my $ret;
    my $rc;
    my $fh = gensym;

    # include both STDOUT and STDERR
    $rc = open($fh, "$cmd 2>&1 |");
    if (! $rc)
    {
        ERROR_OUT("Cannot open pipe for command [%s]: %s", $cmd, $!);
        return undef;
    }

    my @output = <$fh>;

    $rc = close($fh);
    #
    # need to grab $? right after closing the pipe, because logger module
    # will use system command
    #
    $ret = $? >> 8;
    if ($? & 127)
    {
        # command terminated by signal
        ERROR_OUT("Command [%s] terminated by signal %s", $cmd, $? & 127);
        $ret = undef;
    }
    else
    {
        TRACE_OUT(" return: [$ret]");
    }
    if (! $rc)
    {
        #
        # If the file handle came from a piped open, close will additionally
        # return false if one of the other system calls involved fails,
        # or if the program exits with non-zero status.
        # If the only problem was that the program exited non-zero,
        # $! will be set to 0 .
        #
        if ($!)
        {
            WARN_OUT("Cannot close pipe");
        }
    }


    # only log output when loglevel is TRACE
    if (IsTraceOn())
    {
        TRACE_OUT(" command output:");
        foreach my $line (@output)
        {
            my $str = $line;
            chomp $str;
            TRACE_OUT("%s", " | $str");
        }
    }

    wantarray() ? (return ($ret, join("", @output))) : (return $ret);
}

#
# run command with specified timeout
#
# if command doesn't finish in specified time, send SIGTERM and wait for 2
# seconds; if still doesn't finish, send SIGKILL and wait for 2 seconds;
# if still doesn't finish, fail.
#
# in scalar context, return command return value.
# in list context, return command return value and command output.
# If lock seed is specified, a lockfile will be created to ensure
# no other process is running the same command. lockfile name is
# generated from the seed
#
#   $_[0]:  command string
#   $_[1]:  timeout (seconds)
#
#   return: 
#       in scalar context:
#           number  - return value of the command
#           undef   - failed
#       in list context:
#       $1: 
#           number  - return value of the command
#           undef   - failed
#       $2:
#           string  - command output
#
sub _RunCmdWithOpen3($$)
{
    my ($cmd, $timeout) = @_;

    defined($cmd) or return undef;
    defined($timeout) or return undef;

    DEBUG_OUT("Run command [%s] with timeout [%s] seconds", $cmd, $timeout);

    my $PH = gensym;
    my $pid = open3(gensym, $PH, $PH, $cmd);
    DEBUG_OUT("  pid: %s", $pid);

    my $iszombie = 0;
    my ($ret, $error) = _TimedWait($pid, $timeout);

    if ($error)
    {
        DEBUG_OUT("Command [%s] taking too long, sending SIGTERM", $cmd);
        kill(SIGTERM, $pid);
    
        (undef, $error) = _TimedWait($pid, $KILL_TIMEOUT);
    }

    if ($error)
    {
        DEBUG_OUT("Command [%s] is not responding to SIGTERM, sending SIGKILL", $cmd);
        kill(SIGKILL, $pid);
    
        (undef, $error) = _TimedWait($pid, $KILL_TIMEOUT);
    }

    if ($error)
    {
        ERROR_OUT("Command [%s] is not responding to SIGKILL", $cmd);
        $iszombie = 1;
    }

    my @output = <$PH>;

    my $openfail = 0;
    foreach my $line (@output)
    {
        if ($line =~ m/^open3:/g)
        {
            ERROR_OUT("Fail to run command using open3: %s", $line);
            $openfail = 1;
            last;
        }
    }

    if (! $openfail)
    {
        if (! defined($ret))
        {
            if ($iszombie)
            {
                ERROR_OUT("Command [%s] takes too long to run and cannot be killed", $cmd);
            }
            else
            {
                ERROR_OUT("Command [%s] takes too long to run and is killed", $cmd);
            }
        }
        else
        {
            TRACE_OUT(" return: [$ret]");
        }
    }

    my $rc = close($PH);
    if (! $rc)
    {
        ERROR_OUT("Cannot close file handler");
    }

    # only log output when loglevel is TRACE
    if (IsTraceOn())
    {
        TRACE_OUT(" command output:");
        foreach my $line (@output)
        {
            my $str = $line;
            chomp $str;
            TRACE_OUT("%s", " | $str");
        }
    }

    wantarray() ? (return ($ret, join('', @output))) : (return $ret);
}

#
# change file's owner to specified user
#
#   $_[0]:  file
#   $_[1]:  user
#
#   return: 1       - successful
#           undef   - failed
#
sub ChangeOwner($$)
{
    my ($file, $user) = @_;

    if (! defined($file))
    {
        ERROR_OUT("File not defined");
        return undef;
    }
    if (! -e $file)
    {
        ERROR_OUT("File [$file] not exist");
        return undef;
    }

    if (! defined($user))
    {
        ERROR_OUT("User not defined");
        return undef;
    }

    my ($fileuid, $filegid) = (stat $file)[4, 5];
    my ($uid, $gid) = (getpwnam($user))[2, 3];
    if (! defined($uid) || ! defined($gid))
    {
        ERROR_OUT("Cannot get uid/gid of user [$user]");
        return undef;
    }

    if ($uid ne $fileuid || $gid ne $filegid)
    {
        TRACE_OUT("Change owner of [$file] to [$user] $uid:$gid");
        if (! chown($uid, $gid, $file))
        {
            ERROR_OUT("Cannot change ownership of [$file]");
            return undef;
        }
    }

    return 1;
}

#
# create directory and set owner if user is specified.
# if path exists but is not a directory (for example symlink), return failure.
#
#   $_[0]:  dir
#   $_[1]:  user (optional)
#
#   return: 1       - successful
#           undef   - failed
#
sub CreateDir($;$)
{
    my ($dir, $user) = @_;

    if (! defined($dir))
    {
        ERROR_OUT("Dir not defined");
        return undef;
    }

    # if dir not exist, create
    if (! -e $dir)
    {
        TRACE_OUT("create dir [$dir]");
        my $rc;
        eval
        {
            my $old_umask = umask;
            umask 0022;
            $rc = mkpath($dir, 0, 0755);
            umask $old_umask;
        };
        if ($@)
        {
            ERROR_OUT("Cannot create dir [$dir]: $@");
            return undef;
        }
        if (! $rc)
        {
            ERROR_OUT("Cannot create dir [$dir]");
            return undef;
        }
    }
    else
    {
        if (-l $dir)
        {
            # Perl consider a symlink to dir as a dir.
            ERROR_OUT("[$dir] exists but is a symlink");
            return undef;
        }
        if (! -d $dir)
        {
            ERROR_OUT("[$dir] exists but is not a directory");
            return undef;
        }
    }

    # set dir's ownership
    if (defined($user))
    {
        if (! defined(ChangeOwner($dir, $user)))
        {
            ERROR_OUT("Cannot change owner of [$dir]");
            return undef;
        }
    }

    return 1;
}

#
# create empty file and set owner if user is specified. file permission 0644.
# if file exists, only set owner.
#
# default file mode 0644. default owner root.
#
#   $_[0]:  file
#   $_[1]:  options hash reference (optional)
#           supported options are:
#               owner => string, file owner user name
#               group => string, file owner group name
#               mode  => octal number, file mode
#
#   return: 1       - successful
#           undef   - failed
#
sub CreateFile($;$)
{
    my ($file, $options) = @_;

    my $DEFAULT_FILE_MODE = 0644;

    if (! defined($file))
    {
        ERROR_OUT("File not defined");
        return undef;
    }

    my ($has_opt, $opt) = _GetFileOptions($options);
    defined($has_opt) or return undef;

    my ($mode, $uid, $gid);
    
    if ($has_opt)
    {
        defined($opt->{mode}) and $mode = $opt->{mode};
        defined($opt->{uid}) and $uid = $opt->{uid};
        defined($opt->{gid}) and $gid = $opt->{gid};
    }
    defined($mode) or $mode = $DEFAULT_FILE_MODE;
    
    if (-e $file)
    {
        if (-l $file)
        {
            # Perl consider a symlink to file as a regular file.
            ERROR_OUT("[$file] exists but is a symlink");
            return undef;
        }
        if (! -f $file)
        {
            ERROR_OUT("[$file] exists but is not a regular file");
            return undef;
        }
    }
    else
    {
        TRACE_OUT("create empty file [$file] with mode [%04o]", $mode);
        my $fh = gensym;

        my $rc = sysopen($fh, $file, O_RDWR|O_CREAT|O_EXCL, $mode);
        if(! $rc)
        {
            ERROR_OUT("Cannot create file [$file] with mode [%04o]: $!", $mode);
            return undef;
        }

        my $failed;

        if ($IS_CHXXX_SUPPORT_FH)
        {
            # set file uid/gid before close file handler
            # if at least one of uid/gid is specified, use chown to change owner
            if (defined($uid) || defined($gid))
            {
                defined($uid) or $uid = -1;
                defined($gid) or $gid = -1;
                # -1 means chown will keep the original
                DEBUG_OUT("Change file uid:gid to $uid:$gid");
                if (! chown($uid, $gid, $fh))
                {
                    ERROR_OUT("Cannot set ownership for [$file]: $!");
                    $failed = 1;
                }
            }
        }

        $rc = close($fh);
        if (! $rc)
        {
            ERROR_OUT("Cannot close $file: $!");
            $failed = 1;
        }

        # before Perl 5.8.8, need to use file name to chown
        if (! $IS_CHXXX_SUPPORT_FH && ! $failed)
        {
            # set file uid/gid after close file handler
            # if at least one of uid/gid is specified, use chown to change owner
            if (defined($uid) || defined($gid))
            {
                defined($uid) or $uid = -1;
                defined($gid) or $gid = -1;
                # -1 means chown will keep the original
                DEBUG_OUT("Change file uid:gid to $uid:$gid");
                if (! chown($uid, $gid, $file))
                {
                    ERROR_OUT("Cannot set ownership for [$file]: $!");
                    $failed = 1;
                }
            }
        }

        return $failed ? undef : 1;
    }

    return 1;
}

#
# read file into a string, or return undef if file not exist or read failed.
#
#   $_[0]:  filename
#
#   ret:    string  - file content
#           undef   - failed or no such file or file size too big
#
sub ReadFile($)
{
    my $file = $_[0];

    TRACE_OUT("read [$file]");

    if (! -e $file)
    {
        DEBUG_OUT("Read [$file]: file not exist. return undef");
        return undef;
    }
    if (-d $file)
    {
        ERROR_OUT("Cannot read [$file]: $file is a directory");
        return undef;
    }
    if (! -r $file)
    {
        ERROR_OUT("Cannot read [$file]: no permission");
        return undef;
    }

    # if file size exceeds $MAX_FILE_SIZE, don't read.
    my $size = (stat $file)[7];
    if ($size > $MAX_FILE_SIZE)
    {
        ERROR_OUT("$file size exceeds $MAX_FILE_SIZE. Stop reading.");
        return undef;
    }

    my @content_list;
    my $fh = gensym;

    my $rc = sysopen($fh, $file, O_RDONLY|O_NONBLOCK);
    if(! $rc)
    {
        ERROR_OUT("Cannot open file [$file] for read: $!");
        return undef;
    }

    @content_list = <$fh>;

    $rc = close($fh);
    if (! $rc)
    {
        WARN_OUT("Cannot close [$file]: $!");
    }

    return (join "", @content_list);
}

#
# copy file
#
# guarantee the destination file has the same file permission of the source
#
# This function uses lockfile to prevent race condition.
#
#   $_[0]:  source file's name/path
#   $_[1]:  destination file's name/path
#
#   return: 1       - successful
#           undef   - failed
#
sub CopyFile($$)
{
    my ($src, $dst) = @_;

    DEBUG_OUT("Copy from [$src] to [$dst]");

    my ($mode, $uid, $gid);
    if (-e $src)
    {
        ($mode, $uid, $gid) = (stat $src)[2, 4, 5];
        $mode = $mode & 07777;
        TRACE_OUT("Source $src stat:  mode: [%04o]  uid: [%d]  gid: [%d]", $mode, $uid, $gid);
    }
    else
    {
        ERROR_OUT("$src does not exist.");
        return undef;
    }

    if (! -e $dst)
    {
        # create parent dir if not exist
        my $dir = dirname $src;
        if (defined($dir))
        {
            if (! -d $dir)
            {
                TRACE_OUT(" create dir [$dir]");
                eval
                {
                    my $old_umask = umask;
                    umask 0022;
                    mkpath($dir, 0, 0755);
                    umask $old_umask;
                };
                if ($@)
                {
                    ERROR_OUT("Cannot create dir [$dir]: $@");
                    return undef;
                }
            }
        }
    }

    my $lock = CentrifyDC::GP::Lock->new($src);
    if (! defined($lock))
    {
        ERROR_OUT("Cannot obtain lock");
        return undef;
    }
    TRACE_OUT(" lock file: [" . $lock->file() . "]");

    my $rc;
    $rc = copy($src, $dst);
    if (! $rc)
    {
        ERROR_OUT("Cannot copy from $src to $dst: $!");
        return undef;
    }

    # set file attribute
    if (! chmod($mode, $dst))
    {
        ERROR_OUT("Cannot set mode for [$dst]: $!");
        return undef;
    }
    if (! chown($uid, $gid, $dst))
    {
        ERROR_OUT("Cannot set ownership for [$dst]: $!");
        return undef;
    }

    return 1;
}

#
# write string into file
#
# create a temp file, write into it, replace the original file and restore
# file attributes. if options specified, apply to the file.
#
# default file mode 0600. default owner root.
#
# This function uses lockfile to prevent race condition.
#
#   $_[0]:  file name
#   $_[1]:  string
#   $_[2]:  options hash reference (optional)
#           supported options are:
#               owner => string, file owner user name
#               group => string, file owner group name
#               mode  => octal number, file mode
#
#   return: 1       - successful
#           undef   - failed
#
sub WriteFile($$;$)
{
    my ($file, $content, $options) = @_;

    DEBUG_OUT("Write [$file]");

    if (-d $file)
    {
        ERROR_OUT("Cannot write $file: is a directory");
        return undef;
    }

    my ($mode, $uid, $gid);

    # if file exists, get its mode/uid/gid. later they may be overridden by options.
    if (-e $file)
    {
        ($mode, $uid, $gid) = (stat $file)[2, 4, 5];
        $mode = $mode & 07777;
        TRACE_OUT("Backup $file stat:  mode: [%04o]  uid: [%d]  gid: [%d]", $mode, $uid, $gid);
    }

    my ($has_opt, $opt) = _GetFileOptions($options);
    defined($has_opt) or return undef;


    # if has option, get mode/uid/gid from option.
    if ($has_opt)
    {
        defined($opt->{mode}) and $mode = $opt->{mode};
        defined($opt->{uid}) and $uid = $opt->{uid};
        defined($opt->{gid}) and $gid = $opt->{gid};
    }

    # if string size exceeds $MAX_FILE_SIZE, don't write.
    my $size = length($content);
    if ($size > $MAX_FILE_SIZE)
    {
        ERROR_OUT("String size exceeds $MAX_FILE_SIZE. Stop writing.");
        return undef;
    }

    # obtain lock
    my $lock = CentrifyDC::GP::Lock->new($file);
    if (! defined($lock))
    {
        ERROR_OUT("Cannot obtain lock");
        return undef;
    }
    TRACE_OUT(" lock file: [" . $lock->file() . "]");

    # create temp file for write
    my ($fh, $temp_file) = CreateTempFile($file);
    defined($fh) or return undef;
    TRACE_OUT(" temp file: [$temp_file]");

    if (! -e $file)
    {
        # create parent dir if not exist
        my $dir = dirname $file;
        if (defined($dir))
        {
            if (! -d $dir)
            {
                TRACE_OUT(" create dir [$dir]");
                eval
                {
                    my $old_umask = umask;
                    umask 0022;
                    mkpath($dir, 0, 0755);
                    umask $old_umask;
                };
                if ($@)
                {
                    ERROR_OUT("Cannot create dir [$dir]: $@");
                    return undef;
                }
            }
        }
    }

    defined($content) or $content = '';
    my $rc;

    # write file
    $rc = print $fh $content;
    if (! $rc)
    {
        # NOTE: print doesn't fail when file system is full!!!
        ERROR_OUT("Cannot write to $temp_file: $!");
        close($fh);
        if (-e $temp_file)
        {
            unlink $temp_file;
        }
        return undef;
    }

    my $failed;

    if ($IS_CHXXX_SUPPORT_FH)
    {
        # set file mode/uid/gid before close file handler
        if (defined($mode))
        {
            DEBUG_OUT("Change file mode to %04o", $mode);
            if (! chmod($mode, $fh))
            {
                ERROR_OUT("Cannot set mode for [$temp_file]: $!");
                $failed = 1;
            }
        }

        # if at least one of uid/gid is specified, use chown to change owner
        if (defined($uid) || defined($gid))
        {
            defined($uid) or $uid = -1;
            defined($gid) or $gid = -1;
            # -1 means chown will keep the original
            DEBUG_OUT("Change file uid:gid to $uid:$gid");
            if (! chown($uid, $gid, $fh))
            {
                ERROR_OUT("Cannot set ownership for [$temp_file]: $!");
                $failed = 1;
            }
        }
    }

    $rc = close($fh);
    if (! $rc)
    {
        ERROR_OUT("Cannot close $temp_file: $!");
        $failed = 1;
    }

    # before Perl 5.8.8, need to use file name to chown/chmod
    if (! $IS_CHXXX_SUPPORT_FH && ! $failed)
    {
        # set file mode/uid/gid after close file handler
        if (defined($mode))
        {
            DEBUG_OUT("Change file mode to %04o", $mode);
            if (! chmod($mode, $temp_file))
            {
                ERROR_OUT("Cannot set mode for [$temp_file]: $!");
                $failed = 1;
            }
        }

        # if at least one of uid/gid is specified, use chown to change owner
        if (defined($uid) || defined($gid))
        {
            defined($uid) or $uid = -1;
            defined($gid) or $gid = -1;
            # -1 means chown will keep the original
            DEBUG_OUT("Change file uid:gid to $uid:$gid");
            if (! chown($uid, $gid, $temp_file))
            {
                ERROR_OUT("Cannot set ownership for [$temp_file]: $!");
                $failed = 1;
            }
        }
    }

    # if setting mode/uid/gid or close fails, remove temp file and return failure
    if ($failed)
    {
        unlink $temp_file;
        return undef;
    }

    $rc = rename($temp_file, $file);
    if (! $rc)
    {
        ERROR_OUT("Cannot rename $temp_file to $file: $!");
        unlink $temp_file;
        return undef;
    }

    # Restore default context if has selinux
    _RestoreCon($file);

    return 1;
}

#
# parse file options and get mode/uid/gid
#
#   $_[0]:  options hash reference
#           supported options are:
#               owner => string, file owner user name
#               group => string, file owner group name
#               mode  => octal number, file mode
#
#   return: in list context:
#       $1: 
#           number  - 0     - no options
#                     1     - has options and parsed correctly
#                     undef - has options but parse failed
#       $2:
#           hash reference  - hash reference of parsed options
#               uid     => file owner uid
#               gid     => file owner gid
#               mode    => octal number, file mode
#
sub _GetFileOptions($)
{
    my $options = $_[0];

    defined($options) or return (0, undef);

    DEBUG_OUT("Get file options");
 
    if (defined($options))
    {
        my $type = ref($options);
        if ($type ne 'HASH')
        {
            ERROR_OUT("Option must be a hash reference");
            return;
        }
    }

    my ($mode, $uid, $gid);

    # get mode/uid/gid from options
    my $user = $options->{owner};
    if (defined($user))
    {
        DEBUG_OUT("Get uid/gid of user %s", $user);
        ($uid, $gid) = (getpwnam($user))[2, 3];
        if (! defined($uid) || ! defined($uid))
        {
            ERROR_OUT("Cannot get uid/gid of user [$user]");
            return;
        }
        DEBUG_OUT("uid: [%d]  gid: [%d]", $uid, $gid);
    }

    my $group = $options->{group};
    if (defined($group))
    {
        DEBUG_OUT("Get gid of group %s", $group);
        $gid = (getgrnam($group))[2];
        if (! defined($gid))
        {
            ERROR_OUT("Cannot get gid of group [$group]");
            return;
        }
        DEBUG_OUT("gid: [%d]", $gid);
    }

    if (defined($options->{mode}))
    {
        $mode = $options->{mode};
        DEBUG_OUT("Get mode from options:  mode: [%04o]", $mode);
    }

    my $opt = {};
    defined($mode) and $opt->{mode} = $mode;
    defined($uid) and $opt->{uid} = $uid;
    defined($gid) and $opt->{gid} = $gid;

    return (1, $opt);
}

#
# Check if file is empty
# 
#   $_[0]:  file name.  This should not be a symlink.
# 
#   returns:    1     - file is empty
#               0     - file is not empty or doesn't exist
#               undef - file does not exist or if stat fails (eg. no permission)
# 
sub IsFileEmpty($)
{
    my $file = $_[0];
    my $fsize;

    if (-e $file)
    {
        ($fsize) = (stat $file)[7];
        if (defined($fsize))
        {
            if ($fsize == 0)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }
        else
        {
            return undef;
        }
    }
    else
    {
        return undef;
    }
}

#
# Restore default context if has selinux
#
#   $_[0]:  file name
#
#   return: 1       - successful
#           0       - no restorecon
#           undef   - failed
#
sub _RestoreCon($)
{
    my $file = $_[0];

    defined($file) or return undef;

    (-e "/sbin/restorecon") or return 0;

    DEBUG_OUT("Restoring security context of [%s]", $file);
    my $rc = RunCmd(["/sbin/restorecon", $file]);

    if (! defined($rc))
    {
        ERROR_OUT("Fail to restore security context of %s", $file);
        return undef;
    }

    return 1;    
}


#
# get full path of a given file
#
# can't glob the filename directly, because if file doesn't exist, glob
# will fail. to make it work, use getpwnam to get user's home dir
#
# example: ~root/filename
# first glob ~root, then combine the result with /filename
# result will be /var/root/filename
#
#   $_[0]:  filename
#
#   ret:    string  - full path
#           undef   - failed
#
sub GetFullPath($)
{
    my $file = $_[0];

    defined($file) or return undef;

    ($file =~ m|^\~|) or return $file;

    TRACE_OUT("get full path of [$file]");

    my $file_path = $file;
    # extract ~username from file path
    ($file_path =~ s%^\~(.*?)(/.*)%$2%) or return $file;

    my $dir;
    if ($1)
    {
        $dir =(getpwnam($1))[7];
    }
    else
    {
        $dir = $ENV{HOME} || $ENV{LOGDIR} || (getpwuid($>))[7];
    }
    if (! defined($dir))
    {
        ERROR_OUT("Cannot get full path of [$file]");
        return undef;
    }

    $file_path = $dir . $2;

    TRACE_OUT("full path: [$file_path]");

    return $file_path;
}

#
# fix path if not exists or has unexpected permission and ownership
#
# the rationle of checking and fixing temp directory is that, if the temp 
# directory is world writable, attackers can easily insert a symlink using the 
# same file name as the temp file to be written. If the temp file is written by 
# root, then the file in which the symlink is pointed to (e.g. /etc/passwd) 
# will be overwritten. If the temp file has ownership and permission changed by 
# root, then attackers can gain access to root protected files.
#
# therefore, we will use non-world writable temp directory to protect against 
# symlink attack. We will ensure the whole temp directory path is secure. If 
# not, we will try to fix it as described below:
#
# We will check directory per level in path (e.g. /var/centrify/tmp):
#   1. if the directory doesn't exist. Go to (4).
#   2. if this is a symlink, delete it. Go to (4).
#   3. if it exists, check if permission and ownership are OK (*). If OK, 
#      finish. If not, correct permission and ownership as described in (*).
#   4. Create it
#   5. Set permission and ownership as described in (*)
#      
#   (*) Good permission and ownership mean:
#       group=root, owner=root, permission=0755
#
# Exceptions:
#   - we cannot remove /var in Mac because it is a symlink to /private/var by 
#     default. For this reason, we will skip checking symlink for well-known 
#     system pre-defined directory. However, we should still check the 
#     ownership and permission of the destinated directories to ensure the 
#     whole path is secure.
#
# NOTE:
# since now a number of perl scripts will call GetTempDirPath() using root or 
# normal user privileges. Please make sure this subroutine is safe in both 
# cases.
#
#
#   $_[0]:  directory path to be fixed (assume in format /dir1/dir2/dir3)
#   $_[1]:  expected file permission of the directory
#   $_[2]:  expected owner (in uid) of the directory
#   $_[3]:  expected group (in gid) of the directory
#
#   return: 1       - success
#           undef   - failed
#
sub FixPath($$$$)
{
    my ($path, $perm, $uid, $gid) = @_;

    my $ret = 1;
    my $process_mask = 0;
    my $mkdir_flag = 0;
    my $subPath = '';

    my $rc;
    my $file_type;


    # NOTE:
    # Logs are printed in INFO when they are attempting to fix path, and ERROR 
    # when failed to complete the fix. We assume path fixing will not be run 
    # all the times unless there are something weird on customer's machine


    # backup and clear umask
    $process_mask = umask(0);
    
    # remove leading slash for split
    my $index = index($path, '/');
    $path = substr($path, 1) if ($index == 0);
    
    # check directory per level in path
    my @dirs = split('/', $path);
    foreach my $dir (@dirs) 
    {
        $mkdir_flag = 0;

        while (1)
        {
            $subPath = join('/', $subPath, $dir);

            # check if the directory path should be skipped
            if (exists $fix_path_exception_list->{$subPath})
            {
                if ($fix_path_exception_list->{$subPath})
                {
                    TRACE_OUT("Skip fixing [$subPath]...");
                    last;
                }
            }

            # 1. check if file exists
            if (! -e $subPath)
            {
                $mkdir_flag = 1;
                last;
            }

            # 2. check if file should be deleted (e.g. symlink)
            if (-l $subPath)
            {
                INFO_OUT("[$subPath] exists, but is a symlink. Delete.");
                $rc = unlink($subPath);
                if (!$rc)
                {
                    ERROR_OUT("Failed to delete symlink [$subPath]: $!");
                    $ret = undef;
                    last;
                }
                else
                {
                    $mkdir_flag = 1;
                }

                last;
            }
            else
            {
                # we expect file is a directory from now on. if not (e.g. a 
                # plain file), we will leave
                #
                # NOTE:
                # we may need to remove non-directory file if this subroutine is to 
                # use to fix /tmp
                if (! -d $subPath)
                {
                    ERROR_OUT("Cannot fix path as non-directory file is found [$subPath]");
                    $ret = undef;
                    last;
                }
            }

            # 3. check permission and ownership
            # use lstat to ensure symlink is checked
            my $sb = lstat($subPath);
            if (!defined $sb)
            {
                ERROR_OUT("Failed to get permission and ownership [$subPath]");
                $ret = undef;
                last;
            }

            # fix permission
            if (($sb->mode & 0777) != $perm)
            {
                INFO_OUT("Changing permission from %04o to %04o [$subPath]...", ($sb->mode & 0777), $perm);
                $rc = chmod($perm, $subPath);
                if ($rc != 1)
                {
                    ERROR_OUT("Failed to change permission from %04o to %04o [$subPath]", ($sb->mode & 0777), $perm);
                    $ret = undef;
                    last;
                }
            }

            # fix ownership
            if ($sb->uid != $uid || $sb->gid != $gid)
            {
                INFO_OUT("Changing ownership from UID %d GID %d to UID %d GID %d [$subPath]...", $sb->uid, $sb->gid, $uid, $gid);
                $rc = chown($uid, $gid, $subPath);
                if ($rc != 1)
                {
                    ERROR_OUT("Failed to change ownership from UID %d GID %d to UID %d GID %d [$subPath]", $sb->uid, $sb->gid, $uid, $gid);
                    $ret = undef;
                    last;
                }
            }

            # do block is not a real loop in perl, so we use while loop
            # break here to ensure this block will only run once
            last;
        }

        if ($ret && $mkdir_flag eq 1)
        {
            while (1)
            {
                # 4. make directory with correct permission
                # have to be atomic to ensure the directory is created with 
                # expected permission
                INFO_OUT("Creating directory [$subPath]...");
                $rc = mkdir($subPath, $perm);
                if (!$rc)
                {
                    ERROR_OUT("Failed to create directory [$subPath]: $!");
                    $ret = undef;
                    last;
                }

                # 5. set ownership
                INFO_OUT("Setting ownership to UID %d GID %d [$subPath]...", $uid, $gid);
                $rc = chown($uid, $gid, $subPath);
                if ($rc != 1)
                {
                    ERROR_OUT("Failed to set ownership to UID %d GID %d [$subPath]", $uid, $gid);
                    $ret = undef;
                    last;
                }

                # do block is not a real loop in perl, so we use while loop
                # break here to ensure this block will only run once
                last;
            }
        }

        # leave if already failed
        last if (!defined($ret));
    }

    # we need to reset umask before exit
    umask($process_mask);

    return $ret;
}

#
# return temp dir path
#
# If does not need to be non-root writable, we will return CDC's secure temp 
# directory (i.e. /var/centrify/tmp).
#
# If need to be non-root writeable instead, we will return public temp 
# directory (i.e. /tmp).
#
# WARNING:
# We will attempt to fix CDC directories to ensure the path is secure. Beware 
# when call this function with root privilege.
# 
#
#   $_[0]:  whether temp dir needs to be non-root writable
#           (0: not needed, 1: needed)
#
#   return: string  - if found valid temp dir path
#           undef   - if not found or path does not fulfill requirements
#
sub GetTempDirPath($)
{
    my $nonRootWritable = $_[0];

    my $dir = undef;
    my $perm;
    my $uid;
    my $gid;

    if (!$nonRootWritable)
    {
        $dir = $TEMP_DIR_NON_ROOT_READABLE_PATH;
        $perm = $TEMP_DIR_NON_ROOT_READABLE_PERM;
        $uid = $TEMP_DIR_NON_ROOT_READABLE_UID;
        $gid = $TEMP_DIR_NON_ROOT_READABLE_GID;

        $dir = undef if (!FixPath($dir, $perm, $uid, $gid));
    }
    else
    {
        # we assume /tmp has the expected permission and ownership
        $dir = $TEMP_DIR_NON_ROOT_WRITABLE_PATH;
        
        $dir = undef if (! -d $dir);
    }

    defined($dir) or ERROR_OUT("Failed to get temp directory path");
    return $dir;
}

#
# create a temp file and return file handler and filename.
# the filename is generated by appending pid or time to the end of source
# filename.
# default mode is 0600
#
# cannot use File::Temp because it causes weird error on Mac OS X and it's
# not in Perl 5.0
#
#   $_[0]:  seed string
#
#   return:
#       $1: 
#           file handler  - file handler of temp file
#           undef   - failed
#       $2:
#           string  - temp file name
#           undef   - failed
#
sub CreateTempFile($)
{
    my $file = $_[0];

    my $file_temp = undef;
    my $fh = gensym;

    my $tmp;
    my $rc;

    # generate temp file name
    # avoid using pid to make file name less predictable
    foreach (1..$TEMP_FILE_NAME_TRY_COUNT)
    {
        # use random number as suffix
        $tmp = $file . sprintf(".%04x", int(rand(65536)));
        if (! -e $tmp)
        {
            $file_temp = $tmp;
            last;
        }
        else
        {
            # try to remove symlink which conflicts with our temp file
            if (-l $tmp)
            {
                $rc = unlink($tmp);
                if ($rc)
                {
                    $file_temp = $tmp;
                    last;
                }
                else
                {
                    DEBUG_OUT("Cannot delete symlink [$tmp]: $!");
                    next;
                }
            }
            else
            {
                TRACE_OUT("File already exists [$tmp]");
            }
        }
    }

    if (defined($file_temp))
    {
        # NOTE: Setting O_CREAT|O_EXCL prevents the file from being opened if 
        # it is a symbolic link. It does not protect against symbolic links in 
        # the file's path. 
        #
        # NOTE: It is warned that in some UNIX systems, sysopen will fail when 
        # file descriptors exceed a certain value (typically 255) due to the 
        # use of fdopen() C library function.
        #
        # We will use a permission of 0600 because:
        # - We usually create temp file for self use only.
        # - In case we want the temp file to do sth extra, we can chmod to 
        #   relax the permission after creation.
        #
        my $rc = sysopen($fh, $file_temp, O_RDWR|O_CREAT|O_EXCL, 0600);
        if(! $rc)
        {
            ERROR_OUT("Cannot open temp file [$file_temp]: $!");
            return (undef, undef);
        }
        else
        {
            return ($fh, $file_temp);
        }

    }
    else
    {
        ERROR_OUT("Cannot create temp file for [$file]");
        return (undef, undef);
    }
}

#
# traverse symlink and return the actual file name.
#
#   $_[0]:  filename
#
#   ret:    string  - actual file name
#           undef   - failed
#
sub TraverseSymLink($)
{
    my $filename = $_[0];

    (defined($filename) and $filename ne '') or return $filename;

    my @known_links = ();
    push(@known_links, $filename);
    while (defined(my $target = readlink($filename)))
    {
        # if target is a relative path but symlink is absolute path,
        # then we need to add dirname in front of target.
        if ($target =~ m|^[^/]|)
        {
            if ($filename =~ m|^/|)
            {
                my $dir = dirname $filename;
                $target = "$dir/$target";
            }
        }
        # check if target file is in the known link list. this is
        # for preventing self-reference and endless-loop of reference.
        foreach (@known_links)
        {
            if ($target eq $_)
            {
                DEBUG_OUT("self reference found. stop following the symlink.");
                return $filename;
            }
        }
        $filename = $target;
        push(@known_links, $filename);
    }

    return $filename;
}

# <<< SYSTEM <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<



# >>> GENERAL >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# check if two variables are equal.
#
# Both variables can be scalar or reference variables. Supported reference
# types are: SCALAR, ARRAY, and HASH. Other data types are not supported,
# for example, you can't pass hash object to this function.
#
# Two undefined variables are considered equal.
#
#   $_[0]:  variable a
#   $_[1]:  variable b
#   return: 0       - different (or data type unsupported)
#           1       - same
sub IsEqual
{
    my ($a, $b) = @_;

    return _IsEqual($a, $b);
}

# Work function of IsEqual
#   $_[0]:  variable a
#   $_[1]:  variable b
#   @_:     list of parent ref
#   return: 0       - different (or data type unsupported)
#           1       - same
sub _IsEqual
{
    my ($a, $b, @parentref) = @_;

    # if both undefined, then equal
    # if one undefined, then different
    if (defined($a))
    {
        defined($b) or return 0;
    }
    else
    {
        defined($b) ? (return 0) : (return 1);
    }

    my $type = ref($a);

    # different if data type different
    ($type eq ref($b)) or return 0;

    if (! $type)
    {
        ("$a" eq "$b") ? (return 1) : (return 0);
    }
    ("$a" eq "$b") and return 1;

    my %parentrefhash = map { $_, 1 } @parentref;
    if (exists $parentrefhash{$a})
    {
        return 0;
    }
    else
    {
        push(@parentref, $a);
    }

    if ($type eq 'ARRAY')
    {
        # compare number of elements
        (@$a == @$b) or return 0;

        my @arr_a = @$a;
        my @arr_b = @$b;
        while (@arr_a)
        {
            (_IsEqual(pop(@arr_a), pop(@arr_b), @parentref)) or return 0;
        }
    }
    elsif ($type eq 'HASH')
    {
        # compare keys
        my @hashkey_a = sort keys %$a;
        my @hashkey_b = sort keys %$b;
        _IsEqual(\@hashkey_a, \@hashkey_b, @parentref) or return 0;

        # compare values
        foreach my $key (keys %$a)
        {
            _IsEqual($a->{$key}, $b->{$key}, @parentref) or return 0;
        }
    }
    elsif ($type eq 'SCALAR' or $type eq 'REF')
    {
        _IsEqual($$a, $$b, @parentref) or return 0;
    }
    else
    {
        # unsupported reference type
        return 0;
    }

    return 1;
}

#
# check if variable doesn't contain any data.
#
# if variable is undef or is empty array/hash, then it's empty
#
#   $_[0]:  variable
#
#   return: 0       - contains data
#           1       - empty
#
sub IsEmpty($)
{
    my $data = $_[0];

    defined($data) or return 1;

    my $type = ref($data);
    if ($type eq 'ARRAY')
    {
        # at least one element
        (@$data > 0) or return 1;
    }
    elsif ($type eq 'HASH')
    {
        # compare keys
        ((keys %$data) > 0) or return 1;
    }
    elsif ($type eq 'SCALAR')
    {
        return IsEmpty($$data);
    }

    return 0;
}

#
# add elements into an array
# this function will not modify the original array
#
#   $_[0]:  array reference
#   $_[1]:  elements (array reference or scalar)
#
#   return: array reference - successful
#           undef           - failed
#
sub AddElementsIntoArray($$)
{
    my ($array, $elements) = @_;

    if (ref($array) ne 'ARRAY')
    {
        ERROR_OUT("Cannot add elements into array: target is not an array");
        return undef;
    }

    my @new_array = @$array;

    IsEmpty($elements) and return \@new_array;

    my %hash = map { $_, 1 } @$array;
    if (ref($elements))
    {
        if (ref($elements) eq 'ARRAY')
        {
            foreach my $element (@$elements)
            {
                if (! exists($hash{$element}))
                {
                    TRACE_OUT("add [$element] into array");
                    push(@new_array, $element);
                }
            }
        }
        else
        {
            ERROR_OUT("Cannot add elements into array: elements not an array or a scalar");
            return undef;
        }
    }
    else
    {
        if (! exists($hash{$elements}))
        {
            TRACE_OUT("add [$elements] into array");
            push(@new_array, $elements);
        }
    }

    return \@new_array;
}

#
# remove elements from an array
# this function will not modify the original array
#
#   $_[0]:  array reference
#   $_[1]:  elements (array reference or scalar)
#
#   return: array reference - successful
#           undef           - failed
#
sub RemoveElementsFromArray($$)
{
    my ($array, $elements) = @_;

    if (ref($array) ne 'ARRAY')
    {
        ERROR_OUT("Cannot remove elements from array: target is not an array");
        return undef;
    }

    my @new_array = ();

    if (IsEmpty($elements))
    {
        @new_array = @$array;
        return \@new_array;
    }

    my %hash = map { $_, 1 } @$array;
    if (ref($elements))
    {
        if (ref($elements) eq 'ARRAY')
        {
            my @tmp_array = @$array;
            foreach my $element (@$elements)
            {
                # add element only if it's not in the array
                if (exists($hash{$element}))
                {
                    TRACE_OUT("remove [$element] from array");
                    foreach my $array_member (@tmp_array)
                    {
                        IsEqual($array_member, $element) or push(@new_array, $array_member); 
                    }
                    @tmp_array = @new_array;
                    @new_array = ();
                }
            }
            @new_array = @tmp_array;
        }
        else
        {
            ERROR_OUT("Cannot remove elements from array: elements not an array or a scalar");
            return undef;
        }
    }
    else
    {
        if (exists($hash{$elements}))
        {
            TRACE_OUT("remove [$elements] from array");
            foreach my $array_member (@$array)
            {
                IsEqual($array_member, $elements) or push(@new_array, $array_member); 
            }
        }
    }

    return \@new_array;
}

#
# compare two array and generate three lists:
#  1. items only in array 1
#  2. items only in array 2
#  3. items in both array 1 and 2
#
# all members in generated arrays are created in this function, not the
# original members of original arrays.
#
# undef and duplicated entry will be ignored.
#
#   $_[0]:  array reference - array 1
#   $_[1]:  array reference - array 2
#
#   return:
#       $1: 
#           array reference - items only in array 1
#       $2:
#           array reference - items only in array 2
#       $3:
#           array reference - items in both array 1 and 2
#
sub DiffTwoArray($$)
{
    my $array_1 = $_[0];
    my $array_2 = $_[1];

    my @a1_temp = ();
    my @a2_temp = ();

    # remove undef and duplicated entry, then sort
    foreach my $value (@$array_1)
    {
        defined($value) or next;
        push @a1_temp, $value;
    }
    my %hash1 = map { $_, 1 } @a1_temp;

    foreach my $value (@$array_2)
    {
        defined($value) or next;
        push @a2_temp, $value;
    }
    my %hash2 = map { $_, 1 } @a2_temp;

    my @a1 = sort keys %hash1;
    my @a2 = sort keys %hash2;

    my $size_a1 = @a1;
    my $size_a2 = @a2;
    my $p1 = 0;         # pointer to a1
    my $p2 = 0;         # pointer to a2
    my $a1only = [];    # a1 only
    my $a2only = [];    # a2 only
    my $both = [];      # in both a1 and a2
 
    if ($size_a2 == 0)
    {
        $a1only = \@a1;
    }
    elsif ($size_a1 == 0)
    {
        $a2only = \@a2;
    }
    else
    {
        # compare two array
        # put items only in a1 into a1only
        # put items only in a2 into a2only
        for ($p1 = 0, $p2 = 0; $p1 < $size_a1; )
        {
            if ($a1[$p1] lt $a2[$p2])
            {
                push @$a1only, $a1[$p1];
                $p1++;
            }
            elsif ($a1[$p1] gt $a2[$p2])
            {
                push @$a2only, $a2[$p2];
                $p2++;
            }
            else
            {
                push @$both, $a1[$p1];
                $p1++;
                $p2++;
            }

            # reach the end of a2
            if ($p2 >= $size_a2)
            {
                for ( ; $p1 < $size_a1; )
                {
                    push @$a1only, $a1[$p1];
                    $p1++;
                }
                last;
            }

            # reach the end of a1
            if ($p1 >= $size_a1)
            {
                for ( ; $p2 < $size_a2; )
                {
                    push @$a2only, $a2[$p2];
                    $p2++;
                }
                last;
            }
        }
    }

    TRACE_OUT("diff two arrays:");
    TRACE_OUT(" a1:      [" . join(" ", @a1) . "]");
    TRACE_OUT(" a2:      [" . join(" ", @a2) . "]");
    TRACE_OUT(" a1 only: [" . join(" ", @$a1only) . "]");
    TRACE_OUT(" a2 only: [" . join(" ", @$a2only) . "]");
    TRACE_OUT(" both:    [" . join(" ", @$both) . "]");

    return ($a1only, $a2only, $both);
}

sub EscapeSingleQuote($)
{
    my $str = $_[0];
    $str =~ s/'/'\\''/g;
    return $str;
}

# 
# check if current OS supports FIPS. Only Mac, AIX and RedHat Linux supports FIPS.
#
#   return: 1       - FIPS supported
#           0       - FIPS not supported
#           undef   - failed
#
#
sub IsFipsSupported()
{
    my $supported = 0;

    my $platform = CentrifyDC::Platform::GetPlatform();
    if (!defined($platform))
    {
        ERROR_OUT("Failed to detect platform.");
        return undef;
    }
    
    my $os = $platform->getOS();   
    my $os_rev_major = $platform->getOSRevMajor();
    my $os_rev_minor = $platform->getOSRevMinor();
    
    if ($os =~ CentrifyDC::Platform->OS->{RHEL})
    {
        # Note that the OS platforms allowed below are NOT equivalent to the official supported platforms.
        # Since CDC needs to be certified for FIPS 140-2 for each specified platform. 
        # As for now, only RHEL 5, AIX 7.2, Mac 10.6 and 10.7 are officially supported.
        
        if ($os_rev_major >= 5)
        {
            $supported = 1;
        }
        
    }
    elsif ($os =~ CentrifyDC::Platform->OS->{DARWIN})
    {
        $supported = 1;
    }
    elsif ($os =~ CentrifyDC::Platform->OS->{AIX})
    {
        if (($os_rev_major ==7 && $os_rev_minor >= 2) || $os_rev_major > 7)
        {
            $supported = 1;
        }
    }

    TRACE_OUT("fips_supported[$supported]");
    return $supported;
}

sub IsFipsEnabled()
{
    # Check if fips mode is enabed by GP or in centrifydc.conf
    CentrifyDC::GP::Registry::Load(undef);
    my $fips_prop = $CentrifyDC::Config::properties{'fips.mode.enable'};
    my $fips_gp = (CentrifyDC::GP::Registry::Query("machine", "Software/Policies/Centrify/CentrifyDC/Settings/Fips", "current", "fips.mode.enable"))[1];
    my $fips_enabled = ($fips_prop eq "true" or $fips_gp eq "true");
    TRACE_OUT("fips_enabled[$fips_enabled] = fips_prop[$fips_prop] | fips_gp[$fips_gp]");
    return $fips_enabled;
}

# <<< GENERAL <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<



# >>> INIT >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

__InitLogger();

1;
