#!/bin/sh /usr/share/centrifydc/perl/run

################################################################################
#
# Copyright (C) 2015 Centrify Corporation. All rights reserved.
#
# This sample script is used to automate Kerberos setup for Hadoop clusters.
# Centrify DirectControl Agent is required on all cluster nodes.
#
#
# Usage
# -----
#
# To create Kerberos keytabs for Hadoop services:
#   ./kerberos_security_setup.pl --input host-principal-keytab-list.csv --create
#
# To deploy/distribute Kerberos keytabs to cluster nodes:
#   ./kerberos_security_setup.pl --input host-principal-keytab-list.csv --deploy
#
# To undeploy Kerberos keytabs from cluster nodes:
#   ./kerberos_security_setup.pl --input host-principal-keytab-list.csv
#   --undeploy
#
# To delete Kerberos keytabs for Hadoop services:
#   ./kerberos_security_setup.pl --input host-principal-keytab-list.csv --delete
#
# To remove specified default SPN(s) created in computer objects by Centrify agent:
#   ./kerberos_security_setup.pl --input host-principal-keytab-list.csv --remove-spn
#
# To print commands only without executing, use "--dry-run" option. This option
# is useful for review purpose:
#   ./kerberos_security_setup.pl --input host-principal-keytab-list.csv --create
#   --dry-run
#
#
# Troubleshooting
# ---------------
#
# Run commands with "--verbose" option to get more information. Also check
# syslog and centrifydc.log. Enable debug log (using addebug) if needed.
#
#
# Notes
# -----
#
# Root privilege is required to run this script. Please use sudo or dzdo if the
# login user does not have such privilege. Note that login user needs to be
# zone-enabled to use dzdo.
#
# This is strongly recommended to run this script on the master node.
# Hortonworks suggests to create Kerberos keytab files on the node running
# NameNode.
#
# ALL keytab files created are stored in the cluster node running this script.
# If this is a security concern, please remove all keytab files in the backup
# folder manually after "create" command finished. The "delete" command should
# still be able to run without keytab file. Since this script runs adkeytab
# with "--force" option to delete AD objects.
#
#
# Known Issues
# ------------
#
# If the script appears to hang, one possible cause is adkeytab asking for
# confirmation. Re-run this script using "--force" option might work around the
# issue.
#
# Currently only one setting in centrifydc.conf is modified when "deploy"
# command is run. The setting will NOT be restored when "undeploy" command is
# run. Please restore the setting manually if needed.
#
# Similarly, the http principal in machine account on Active Directory is
# deleted when "deploy" command is run. The principal will NOT be restored when
# "undeploy" command is run. Please restore the principal manually (using
# adkeytab --addspn) if needed.
#
# The script will run commands on all cluster nodes via SSH. If the SSH user is
# non-root, some commands might fail due to insufficient privilege. Root
# privilege can be authorized using sudo or dzdo, and such setting can be
# configured in hadoop.conf.
#
# The umask of the local privilege user and SSH user are NOT checked. Please
# make sure the umask is protective (e.g. 0022).
#
#
# Revision
# --------
#
# $Change: 319962 $
# $Revision: #1 $
# $DateTime: 2018/02/15 08:56:48 $
#
################################################################################

use strict;
use warnings;

use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;

use Getopt::Long;
use File::Basename;

use CentrifyDC::GP::General qw(RunCommand);

use lib '/usr/share/centrifydc/samples/hadoop/perl';
use CentrifyDC::Hadoop::Logger qw(:debug_user :debug_verbose);
use CentrifyDC::Hadoop::Config qw(LoadUserConfig);
use CentrifyDC::Hadoop::General qw(:common :create :delete :deploy :undeploy);

# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $CENTRIFY_INSTALL_PATH = "/usr/share/centrifydc";
my $HADOOP_CONFIG_PATH = "$CENTRIFY_INSTALL_PATH/samples/hadoop/hadoop.conf";

#
# Executables
#

my $ADKEYTAB = "/usr/sbin/adkeytab";

#
# Global variables
#

my %opts;

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

sub Usage()
{
    print <<EOF;
usage: $0 [commands] -i <ambari.csv>

    Commands:
        -C, --create            Create Hadoop service accounts and Kerberos keytabs.
        -D, --delete            Delete Hadoop service accounts and Kerberos keytabs.
        -P, --deploy            Distribute Kerberos keytabs to cluster nodes.
        -Q, --undeploy          Clean up Kerberos keytabs from cluster nodes.
        -R, --remove-spn        Remove SPN(s) from computer accounts and centrifydc.conf.

    Parameters:
        -i, --input             CSV file generated by Ambari listing Hadoop service accounts and Kerberos tickets.
        -c, --conf              Configuration file to load (default: $HADOOP_CONFIG_PATH).
        -d, --dry-run           Print all commands and exit without executing.
        -f, --force             Force create/delete Hadoop service accounts and files.
        -x, --ignore-error      Ignore error during execution.
        -V, --verbose           Print debugging information for each operation.
        -v, --version           Print version information and exit.
        -h, --help              Print this help information and exit.
EOF

    exit(1);
}

sub DoUndeploy($)
{
    my $input_file = $_[0];

    my $rc = 1;

    # Root privilege is required. Since the keytab directory is root owned on
    # all cluster nodes.
    $rc = CheckRootPrivilege();
    return $rc if ($rc);

    my ($services, $cluster) = ReadCSV($input_file);
    return $rc if (!defined($services));

    # Skip this function call if the cluster nodes are retired and disconnected
    # from AD.
    $rc = CheckClusterCDCStatus($cluster);
    return $rc if ($rc);

    $rc = DeleteClusterKeytabs($services);
    return $rc if ($rc);

    $rc = DeleteClusterKeytabDirs($cluster);
    return $rc if ($rc);

    return $rc;
}

sub DoDeploy($)
{
    my $input_file = $_[0];

    my $rc = 1;

    # Root privilege is required. Since the keytab directory is root owned on
    # all cluster nodes.
    $rc = CheckRootPrivilege();
    return $rc if ($rc);

    my ($services, $cluster) = ReadCSV($input_file);
    return $rc if (!defined($services));

    $rc = CheckLocalKeytabs($services);
    return $rc if ($rc);

    $rc = CheckClusterCDCStatus($cluster);
    return $rc if ($rc);

    $rc = CreateClusterKeytabDirs($cluster);
    return $rc if ($rc);

    $rc = TransferKeytabs($services);
    return $rc if ($rc);

    $rc = ChownKeytabs($services);
    return $rc if ($rc);

    $rc = ChmodKeytabs($services);
    return $rc if ($rc);

    return $rc;
}

sub DoDelete($)
{
    my $input_file = $_[0];

    my $rc = 1;

    $rc = CheckRootPrivilege();
    return $rc if ($rc);

    $rc = CheckValidTGT();
    return $rc if ($rc);

    my ($services, $cluster) = ReadCSV($input_file);
    return $rc if (!defined($services));

    # Currently we delete AD objects using local keytabs. So only local
    # adclient is required to be AD connected.
    $rc = CheckLocalCDCStatus();
    return $rc if ($rc);

    $rc = DeleteKeytabs($services);
    return $rc if ($rc);

    $rc = DeleteLocalKeytabDirs($cluster);
    return $rc if ($rc);

    return $rc;
}


sub DoCreate($)
{
    my $input_file = $_[0];

    my $rc = 1;

    $rc = CheckRootPrivilege();
    return $rc if ($rc);

    $rc = CheckValidTGT();
    return $rc if ($rc);

    my ($services, $cluster) = ReadCSV($input_file);
    return $rc if (!defined($services));

    # Before creating per-host http principals, we need to remove http SPN from
    # AD computer objects. Therefore, we still need to remote execute commands
    # from the master node for the centralized automation script.

    $rc = CheckClusterCDCStatus($cluster);
    return $rc if ($rc);

    $rc = DeleteSPN($cluster);
    return $rc if ($rc);

    $rc = ModifyAdclientConfig($cluster);
    return $rc if ($rc);

    # From this onwards all operations are local

    $rc = CreateLocalKeytabDirs($cluster);
    return $rc if ($rc);

    $rc = CreateKeytabs($services);
    return $rc if ($rc);

    # TODO: store csv to tmp keytab dir (so can compare when cluster nodes are
    # added/retired)

    return $rc;
}

sub DoRemoveSPN($)
{
    my $input_file = $_[0];

    my $rc = 1;

    $rc = CheckRootPrivilege();
    return $rc if ($rc);

    $rc = CheckValidTGT();
    return $rc if ($rc);

    my ($services, $cluster) = ReadCSV($input_file);
    return $rc if (!defined($services));

    # Before creating per-host http principals, we need to remove http SPN from
    # AD computer objects. Therefore, we still need to remote execute commands
    # from the master node for the centralized automation script.

    $rc = CheckClusterCDCStatus($cluster);
    return $rc if ($rc);

    $rc = DeleteSPN($cluster);
    return $rc if ($rc);

    $rc = ModifyAdclientConfig($cluster);
    return $rc if ($rc);

    return $rc;
}

sub DoVersion()
{
    my $command = "$ADKEYTAB --version";

    my ($rc, $command_output) = RunCommand($command);
    if ($rc)
    {
        my $msg = "Failed to get CentrifyDC version: $rc";
        USER_ERROR($msg);
        return $rc;
    }

    print("$command_output");
    return 0;
}

sub GetUserOptions()
{
    #
    # Set options from parameters
    #
    
    $CentrifyDC::Hadoop::General::OPT_DRY_RUN = $opts{d} ? 1 : 0;
    $CentrifyDC::Hadoop::General::OPT_FORCE = $opts{f} ? 1 : 0;
    $CentrifyDC::Hadoop::General::OPT_IGNORE_ERROR = $opts{x} ? 1 : 0;
    $CentrifyDC::Hadoop::Logger::OPT_VERBOSE = $opts{V} ? 1 : 0;

    my $conf_file = $HADOOP_CONFIG_PATH;
    $conf_file = $opts{c} if $opts{c};

    #
    # Print parameters for debugging
    #

    VERBOSE_DEBUG("=== Optional parameters ===");

    VERBOSE_DEBUG("Dry run: $CentrifyDC::Hadoop::General::OPT_DRY_RUN");
    VERBOSE_DEBUG("Force: $CentrifyDC::Hadoop::General::OPT_FORCE");
    VERBOSE_DEBUG("Ignore error: $CentrifyDC::Hadoop::General::OPT_IGNORE_ERROR");
    VERBOSE_DEBUG("Verbose: $CentrifyDC::Hadoop::Logger::OPT_VERBOSE");
    VERBOSE_DEBUG("Config file path: $conf_file");

    VERBOSE_DEBUG("=== End of optional parameters ===");

    return $conf_file;
}

# Note that we do not check if user has the privilege to read the config file.
# The file owner and permission will do this for us.
sub GetUserInput()
{
    #
    # Get user options from input parameters. Note that --verbose does not work
    # until this function is run.
    #

    my $conf_file = GetUserOptions();

    #
    # Load user options from config file
    #

    if (! -f $conf_file)
    {
        USER_ERROR("Failed to find config file $conf_file");
        return 1;
    }

    my $success = LoadUserConfig($conf_file);
    if (!$success)
    {
        USER_ERROR("Failed to load config file $conf_file");
        return 1;
    }

    return 0;
}

sub Run()
{
    my $rc = 1;

    #
    # Dispatch actions
    #

    # Actions which do not require user input
    if ($opts{v})
    {
        $rc = DoVersion();
        return $rc;
    }
    elsif ($opts{h})
    {
        Usage();
        return 0;
    }

    # Check if all mandatory input parameters have been provided
    my $input_file = $opts{i};
    if (!$input_file)
    {
        USER_ERROR("Input file not specified");
        Usage();
        return 1;
    }

    # Get user input
    $rc = GetUserInput();
    return $rc if ($rc);

    # Actions which require user input
    if ($opts{C})
    {
        $rc = DoCreate($input_file);
    }
    elsif ($opts{D})
    {
        $rc = DoDelete($input_file);
    }
    elsif ($opts{P})
    {
        $rc = DoDeploy($input_file);
    }
    elsif ($opts{Q})
    {
        $rc = DoUndeploy($input_file);
    }
    elsif ($opts{R})
    {
        $rc = DoRemoveSPN($input_file);
    }
    else
    {
        Usage();
    }

    return $rc;
}




# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# Get options
#

Getopt::Long::Configure("bundling", "no_ignore_case");

GetOptions( "C|create" => \$opts{C},
            "D|delete" => \$opts{D},
            "P|deploy" => \$opts{P},
            "Q|undeploy" => \$opts{Q},
            "R|remove-spn" => \$opts{R},
            "i|input=s" => \$opts{i},
            "c|conf=s" => \$opts{c},
            "d|dry-run" => \$opts{d},
            "f|force" => \$opts{f},
            "x|ignore-error" => \$opts{x},
            "V|verbose" => \$opts{V},
            "v|version" => \$opts{v},
            "h|help" => \$opts{h} ) or Usage();

#
# Perform actions
#

my $rc = Run();

exit($rc);
