################################################################################
#
# Copyright (C) 2015 Centrify Corporation. All rights reserved.
#
#
# General subroutines for Hadoop sample scripts.
#
# Revision
# --------
#
# $Change: 319962 $
# $Revision: #1 $
# $DateTime: 2018/02/15 08:56:48 $
#
################################################################################

package CentrifyDC::Hadoop::General;

use strict;
use warnings;

use CentrifyDC::GP::General qw(RunCommand ReadFile);
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);

use File::Spec;

use lib '/usr/share/centrifydc/samples/hadoop/perl';
use CentrifyDC::Hadoop::Logger qw(:debug_user :debug_verbose);

# >>> EXPORT >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

use base 'Exporter';

#our @EXPORT_OK = qw();

our %EXPORT_TAGS = (
    'common'        => [qw(CheckRootPrivilege CheckValidTGT ReadCSV CheckClusterCDCStatus CheckLocalCDCStatus DeleteSPN ModifyAdclientConfig)],
    'create'        => [qw(CreateLocalKeytabDirs CreateKeytabs)],
    'delete'        => [qw(DeleteKeytabs DeleteLocalKeytabDirs)],
    'deploy'        => [qw(CheckLocalKeytabs CreateClusterKeytabDirs TransferKeytabs ChownKeytabs ChmodKeytabs)],
    'undeploy'      => [qw(DeleteClusterKeytabs DeleteClusterKeytabDirs)],
);

Exporter::export_ok_tags('common');
Exporter::export_ok_tags('create');
Exporter::export_ok_tags('delete');
Exporter::export_ok_tags('deploy');
Exporter::export_ok_tags('undeploy');

# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $CENTRIFY_INSTALL_PATH = $CIMS_INSTALL;
my $CENTRIFY_KRB5_BIN_PATH = "$CENTRIFY_INSTALL_PATH/kerberos/bin";

my $ADKEYTAB = "${CIMS_USR}/sbin/adkeytab";
my $ADPROPERTY = "${CIMS_INSTALL}/bin/adproperty";
# Use this line to separate SSH command output from SSH banner info. Since we
# don't want SSH banner info from output of RunCommand(). We choose a simpler
# approach instead of changing behavior of RunCommand().
my $SSH_BANNER_PATTERN = "__________HADOOP_SSH_BANNER_SEPARATOR__________";
my $PROP_CDC_KEY_KRB5_SERVICE_PRINCIPALS = 'adclient.krb5.service.principals'; 
my $ERR_ADPROP_NOT_FOUND = 156;

my $LABEL_AMBARI_V1_6_1 = 'Ambari v1.6.1'; 
my $VALUE_AMBARI_V1_6_1 = 8; 
my $LABEL_AMBARI_V2_1_2 = 'Ambari v2.1.2'; 
my $VALUE_AMBARI_V2_1_2 = 12; 

#
# Configurable variables
#

our $OPT_DRY_RUN        = 0         unless defined $OPT_DRY_RUN;
our $OPT_FORCE          = 0         unless defined $OPT_FORCE;
our $OPT_IGNORE_ERROR   = 0         unless defined $OPT_IGNORE_ERROR;

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

sub GetShortHostName($$)
{
    my $hostname = $_[0];
    my $length = $_[1];

    my ($short_name) = split(/\./, $hostname);
    if (defined($length) && $length > 0)
    {
        if (length($short_name) > $length)
        {
            # get the last n chars
            $short_name = substr($short_name, length($short_name) - $length, $length);
        }
    }

    return $short_name;
}

#
# Determine if the account is headless or per-host. Headless accounts are
# shared among all nodes in the Hadoop cluster. Per-host accounts, on the
# contrary, are for Hadoop services/roles dedicated on a single node.
#
# This function determines if the account is headless using the Kerberos
# principal format. Traditionally, a principal is divided into three parts: the
# primary, the instance, and the realm. The format of a typical Kerberos V5
# principal is primary/instance@REALM.
#
# 1. Per-host accounts are using principals in format primary/<fqdn>@REALM
# where fqdn is the fully qualified domain name of the host.
#
# 2. Headless accounts are using principals in format primary@REALM or
# primary/instance@REALM where instance != fqdn. 
# 
# On Hortonworks and Cloudera, a headless account takes format primary@REALM by 
# default, e.g. hdfs@EXAMPLE.COM. However, on MapR it is suggested to 
# take the format primary/<cluster name>@REALM as headless account name, e.g. 
# mapr/my.cluster.com@EXAMPLE.COM.
#
# The above observation holds true on Cloudera, Hortonworks, and MapR
# distributions.
#
#  $_[0]:   Principal name
#  $_[1]:   Host name list
#
# Return:   1 - headless account
#           0 - per-host account
#
sub IsHeadlessAccount($$)
{
    my $upn = $_[0];
    my $hosts = $_[1];
    my $primary; 
    my $instance;
    my $realm;
    
    if (keys %$hosts > 1)
    {
        # shared among more than 1 node, must be headless
        return 1;
    }
    
    # should be only 1 node in %$hosts here
    my ($host) = sort keys %$hosts;

    if ($upn =~ /(.+)\/(.+)@(.+)/)
    {
         return 0 if ($2 eq $host);
    }

    return 1;
}


#
# Determine if the principal is a SPN style principals in Active Diretory.
# Related hadoop bugs: https://issues.apache.org/jira/browse/HADOOP-10183
#
# A SPN style principal is in format primary/instance@REALM, 
# e.g. HTTP/hwc1-h1.example.com@EXAMPLE.COM. If it is a SPN style principal,
# "--upn" and "--principal" options can be used in adkeytab command.
#
# A pricipal can also be an UPN style principals, which is in format 
# primary@REALM, e.g. hdfs@EXAMPLE.COM. In this case, "--upn" and
# "--principal" options should be absent.
#
#  $_[0]:   Principal name
#
# Return:   1 - SPN style
#           0 - UPN style
#
sub IsSpnStylePrincipal($)
{
    my $principal = $_[0];

    if ($principal =~ /(.*)\/(.*)@(.*)/)
    {
        return 1;
    }

    return 0;

}

#
# Get the name of service accounts based on their principals. Feel free to
# modify this function to get the desired account names.
#
#  $_[0]:   Principal name
#
# Return:   Service account name in SAM
#
sub GetServiceAccountName($)
{
    my $principal = $_[0];
    my $primary = "";
    my $instance = "";
    my $realm = "";
    my @w;
    my $acct = "";

    if (!IsSpnStylePrincipal($principal))
    {
        $principal =~ /(.+)@(.+)/;
        ($primary, $realm) = ($1, $2);
        $acct = $primary;
    }
    else
    {
        $principal =~ /(.+)\/(.+)@(.+)/;
        ($primary, $instance, $realm) = ($1, $2, $3);  
        # Per host accounts
        my $prefix = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_CLUSTER_SHORTNAME};
        my $length = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_HOST_SHORTNAME_EXTRACTED_LAST_CHARS};
        my $is_cluster_prefix_required = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_CLUSTER_SHORTNAME_ENABLE};
        $acct = $prefix . '-' if ($is_cluster_prefix_required);
        $acct .= $primary . '-' . GetShortHostName($instance, $length);
    }

    # TODO: Check max length. See hadoop.host.shortname.extracted.last.chars.
    
    return $acct;
}

sub GetKeytabName($$$$)
{
    my $hosts = $_[0];
    my $upn = $_[1];
    my $keytab_file = $_[2];
    my $keytab_dir = $_[3];  
    my $keytab;
    my $prefix;

    if (IsHeadlessAccount($upn, $hosts))
    {
        $keytab = File::Spec->catpath("", $keytab_dir, $keytab_file);
    }
    else
    {
        # Should have only one host if not headless accounts. So take the first element from hash.
        my ($host) = sort keys %$hosts;

        $prefix = GetShortHostName($host, 0);
        $keytab = File::Spec->catpath("", $keytab_dir, $prefix);
        $keytab = File::Spec->catpath("", $keytab, $keytab_file);
    }

    return $keytab;
}

#
# Note: Need to take care of SSH banner when getting SSH command output. SSH
#       banner info will show in stderr, but sub RunCommand() will return both 
#       stderr and stdout as command output.
#       We could use $SSH_BANNER_PATTERN to deal with the banner.
#       Others like /etc/issue, /etc/motd will affect SSH command output only if
#       they are configured as SSH banner.
#
sub GetSSHCommand($$)
{
    my $host = $_[0];
    my $command = $_[1];

    # Set KRB5CCNAME before running ssh (optional). The file should be owned by
    # root.
    # e.g. my $krb5ccname = "/tmp/krb5cc_hadoop_ssh";
    my $krb5ccname = "";
    $krb5ccname = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_SHELL_KRB5CCNAME} if ($CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_SHELL_KRB5CCNAME_ENABLE});

    #
    # The option "requiretty" is enabled by default on sudoers file. So we have
    # to run like this:
    # "ssh -t hostname sudo <cmd>"
    #
    # There is a tradeoff, however, is that every output returned by SSH will
    # look like this:
    # ...
    # Connection to <host> closed.
    #
    # Such output is suggested to be avoidable by adding "-o LogLevel=QUIET" or
    # simply "ssh -q". But other messages will also get suppressed. So we leave
    # this to end users to decide.
    #
    my $ssh_command = "ssh -t ";

    $ssh_command = "KRB5CCNAME=$krb5ccname $ssh_command" if $krb5ccname;
    $ssh_command = $ssh_command . "$CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_SHELL_USER}\@" if $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_SHELL_USER_ENABLE};
    $ssh_command = $ssh_command . "$host \"";
    $ssh_command = $ssh_command . "$CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_SHELL_PRIVILEGE} " if $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_SHELL_PRIVILEGE_ENABLE};
    $ssh_command = $ssh_command . "$command\"";

    return $ssh_command;
}

sub GetSCPCommand($$$)
{
    my $host = $_[0];
    my $src = $_[1];
    my $dest = $_[2];

    # Set KRB5CCNAME before running scp (optional). The file should be owned by
    # root.
    # e.g. my $krb5ccname = "/tmp/krb5cc_hadoop_scp";
    my $krb5ccname = "";
    $krb5ccname = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_COPY_KRB5CCNAME} if ($CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_COPY_KRB5CCNAME_ENABLE});

    # Get SCP command
    my $scp_command = "scp $src ";

    $scp_command = "KRB5CCNAME=$krb5ccname $scp_command" if $krb5ccname;
    $scp_command = $scp_command . "$CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_COPY_USER}\@" if $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SECURE_COPY_USER_ENABLE};
    $scp_command = $scp_command . "$host:$dest";

    return $scp_command;
}

#
# Read CSV file which contains Kerberos keytab information.
#
# Two hashes are returned. One is per service. Another is per cluster node.
#
# Note this is possible that the CSV file only has host names defined and
# without any Kerberos keytab information, e.g.:
#
#   host1.example.com,,,,,,,
#   host2.example.com,,,,,,,
#
# Since Cloudera Manager automatically distributes all per host keytab files,
# only the hdfs@REALM superuser principal is required. If customers opt to
# create the AD service account by themselves, then this script will only
# modify centrifydc.conf and remove http principal from machine account on AD
# when "create" command is run.
#
sub ReadCSV($)
{
    my $fn = $_[0];
    my $services = {};
    my $cluster = {};

    # Line start from 1
    my $line_num = 0;

    my $file_content = ReadFile($fn);
    if (!defined($file_content))
    {
        USER_ERROR("Failed to open file $fn to read");
        return (undef, undef);
    }

    my @lines = split(/\n/, $file_content);
    my @fields;
    my $columns;

    foreach my $line (@lines)
    {
        $line_num++;
        next if ($line =~ /^#/);

        chomp($line);

        # Headless accounts:
        #   ambari-qa@EXAMPLE.COM
        #   hdfs@EXAMPLE.COM
        #   hbase@EXAMPLE.COM
        
        # Per host accounts:
        #   HTTP/_HOST@EXAMPLE.COM
        #   dn/_HOST@EXAMPLE.COM
        #   hbase/_HOST@EXAMPLE.COM

        # CSV format:
        # (1) with Ambari v1.6.1 (no header line) - 8 fields
        #   host1.example.com,NameNode,nn/host1.example.com@EXAMPLE.COM,nn.service.keytab,
        #     /etc/security/keytabs,hdfs,hadoop,400
        # (2) with Ambari v2.1.2 (with header line) - 12 fields
        #   host,description,principal name,principal type,local username,
        #     keytab file path,keytab file owner,keytab file owner access,keytab file group,keytab file group access,
        #     keytab file mode,keytab file installed
        #   host1.example.com,namenode_nn,nn/host1.example.com@EXAMPLE.COM,SERVICE,hdfs,
        #     /etc/security/keytabs/nn.service.keytab,hdfs,r,hadoop,,
        #     400,unknown

        my $host;
        my $user_desc;
        my $principal_name;
        my $keytab_file;
        my $keytab_dir;

        my $user;
        my $group;
        my $perm;
        my $principle_type; # Do we need this?
        my $local_username; # Not use

        my $keytab_file_path; # Need to parse back to $keytab_file and $keytab_dir
        my $user_mode; # Not use
        my $group_mode; # Not use
        my $install_status; # Not use
        my $volume; # Temporary buffer - Not use
        
        @fields = split(",", $line);
        $columns = scalar @fields;

        if ($line_num == 1)
        {
            my $format = "Unexpected";
            
            if ($columns == $VALUE_AMBARI_V1_6_1)
            {
                $format = $LABEL_AMBARI_V1_6_1;
            }
            elsif ($columns == $VALUE_AMBARI_V2_1_2)
            {
                $format = $LABEL_AMBARI_V2_1_2;
            }
            VERBOSE_INFO("Number of fields on line $line_num of $fn is $columns - Treat it as $format CSV format.");
        }

        if ($columns == $VALUE_AMBARI_V1_6_1)
        {
            ($host, $user_desc, $principal_name, $keytab_file, $keytab_dir, $user, $group, $perm) = split(",", $line);
        }
        elsif ($columns == $VALUE_AMBARI_V2_1_2)
        {
            ($host, $user_desc, $principal_name, $principle_type, $local_username, $keytab_file_path, $user, $user_mode, $group, $group_mode, $perm, $install_status) = split(",", $line);
    
            #
            # Remove the header line if exist
            #
    
            if ($keytab_file_path eq "keytab file path" and $perm eq "keytab file mode")
            {
                VERBOSE_INFO("Skipped line $line_num from $fn. It is a header.");
                next;
            }
    
            #
            # Split $keytab_file_path into $keytab_dir and $keytab_file
            #
            
            ($volume, $keytab_dir, $keytab_file) = File::Spec->splitpath($keytab_file_path);
        }
        else
        {
            USER_ERROR("Number of fields is neither $VALUE_AMBARI_V1_6_1 nor $VALUE_AMBARI_V2_1_2 in line $line_num from $fn. Unexpected CSV format.");
            return (undef, undef);
        }

        if (!defined($perm))
        {
            USER_WARN("Failed to parse line $line_num from $fn. Unexpected format.");
            next;
        }

        if ($host eq "")
        {
            USER_ERROR("Hostname not specified in line $line_num from $fn. Unexpected format.");
            return (undef, undef);
        }

        #
        # Prepare service list
        #

        if ($principal_name ne "")
        {
            if (!defined($services->{$principal_name}))
            {
                my $service_hosts = {};
                $service_hosts->{$host} = 1;

                my $service = {};
                $service->{'hosts'} = $service_hosts;
                $service->{'user_desc'} = $user_desc;
                $service->{'principal_name'} = $principal_name;
                $service->{'keytab_file'} = $keytab_file;
                $service->{'keytab_dir'} = $keytab_dir;
                $service->{'user'} = $user;
                $service->{'group'} = $group;
                $service->{'perm'} = $perm;

                $services->{$principal_name} = $service;
            }
            else
            {
                my $service = $services->{$principal_name};

                # Fields should be the same except hostname (headless accounts) and
                # user description (per-host accounts).
                if ($service->{'principal_name'} ne $principal_name ||
                    $service->{'keytab_file'} ne $keytab_file ||
                    $service->{'keytab_dir'} ne $keytab_dir ||
                    $service->{'user'} ne $user ||
                    $service->{'group'} ne $group ||
                    $service->{'perm'} ne $perm)
                {
                    USER_WARN("Duplicated record of service $principal_name with unexpected fields. Skip this record.");
                    next;
                }

                # Check if hostname has been recorded. Normally per-host accounts
                # (like HTTP) are used by multiple service in one cluster node. On
                # the other hand, headless accounts are only used once per cluster
                # node.
                my $service_hosts = $service->{'hosts'};
                if (defined($service_hosts->{$host}))
                {
                    $service_hosts->{$host}++;
                }
                else
                {
                    $service_hosts->{$host} = 1;
                }
            }
        }

        #
        # Prepare cluster node list
        #

        if (!defined($cluster->{$host}))
        {
            my $node_services = {};
            $node_services->{$principal_name} = 1 if ($principal_name ne "");

            my $node_keytab_dirs = {};
            $node_keytab_dirs->{$keytab_dir} = 1 if ($keytab_dir ne "");

            my $node = {};
            $node->{'host'} = $host;
            $node->{'services'} = $node_services;
            $node->{'keytab_dirs'} = $node_keytab_dirs;

            # Since all keytab files and directories over the cluster should
            # belong to the same group (default "hadoop"). Here we assume only
            # one group is defined within each cluster node.
            $node->{'group'} = $group;

            $cluster->{$host} = $node;
        }
        else
        {
            my $node = $cluster->{$host};

            my $node_services = $node->{'services'};
            if (!defined($node_services->{$principal_name}))
            {
                $node_services->{$principal_name} = 1 if ($principal_name ne "");
            }

            my $node_keytab_dirs = $node->{'keytab_dirs'};
            if (!defined($node_keytab_dirs->{$keytab_dir}))
            {
                $node_keytab_dirs->{$keytab_dir} = 1 if ($keytab_dir ne "");
            }

            if ($node->{'group'} ne $group)
            {
                USER_ERROR("Unexpected group $group from CSV file (expected: $node->{'group'}) for host $host. Cannot continue.");
                return (undef, undef);
            }
        }
    }

    return ($services, $cluster);
}

sub DeleteKeytabs($)
{
    my $services = $_[0];

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $principal_name = $service->{'principal_name'};
        my $keytab_file = $service->{'keytab_file'};
        my $keytab_dir = $service->{'keytab_dir'};

        my $acct_name = GetServiceAccountName($principal_name);
        my $keytab = GetKeytabName($hosts, $principal_name, $keytab_file, $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR});

        # Warning:
        # This script might hang if '-f' option is not supplied to adkeytab.
        # Since adkeytab will ask proceed or not if some account components not
        # found. For example, keytab file or AD account not found. So we force
        # delete by default.
        my $command = "$ADKEYTAB --delete";
        $command = "$command --force";
        $command = "$command --keytab $keytab";
        $command = "$command $acct_name";

        if ($OPT_DRY_RUN)
        {
            print "$command\n";
            next;
        }

        # Check keytab exists or not
        if ((! -f $keytab) && !$OPT_FORCE)
        {
            my $msg = "Kerberos keytab $keytab cannot be found. Please use '--force' option to force delete AD object.";
            USER_ERROR($msg);
            return 1 unless $OPT_IGNORE_ERROR;
        }

        # Run adkeytab
        USER_INFO("Deleting Kerberos keytab $keytab and AD service account $acct_name...");
        my ($rc, $command_output) = RunCommand($command);

        VERBOSE_INFO($command_output) if ($command_output);
        if ($rc)
        {
            my $msg = "Failed to delete Kerberos keytab: $rc";
            USER_ERROR($msg);
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    return 0;
}

sub CreateKeytabs($)
{
    my $services = $_[0];
    my $ou = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_SERVICE_CONTAINER};

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $principal_name = $service->{'principal_name'};
        my $keytab_file = $service->{'keytab_file'};
        my $keytab_dir = $service->{'keytab_dir'};
        my $user = $service->{'user'};
        my $group = $service->{'group'};

        my $acct_name = GetServiceAccountName($principal_name);
        my $keytab = GetKeytabName($hosts, $principal_name, $keytab_file, $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR});
        my $is_spn_style = IsSpnStylePrincipal($principal_name);

        my $command = "$ADKEYTAB --new";
        
        $command = "$command --force" if ($OPT_FORCE);
        $command = "$command --computer-object" if ($CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_USE_COMP_OBJECT});
        $command = "$command --upn $principal_name" if ($is_spn_style || $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_USE_COMP_OBJECT});
        $command = "$command --principal $principal_name" if ($is_spn_style);
        $command = "$command --password-never-expire";
        $command = "$command --keytab $keytab";
        $command = "$command --container \"$ou\"";
        $command = "$command $acct_name";

        if ($OPT_DRY_RUN)
        {
            print "$command\n";
            next;
        }

        # run adkeytab
        USER_INFO("Creating Kerberos keytab $keytab and AD service account $acct_name...");
        my ($rc, $command_output) = RunCommand($command);

        VERBOSE_INFO($command_output) if ($command_output);
        if ($rc)
        {
            my $msg = "Failed to create Kerberos keytab: $rc";
            USER_ERROR($msg);
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    return 0;
}

sub CheckLocalKeytabs($)
{
    my $services = $_[0];

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $principal_name = $service->{'principal_name'};
        my $keytab_file = $service->{'keytab_file'};

        # Should have only one host if not headless accounts. So we can simply
        # check if this keytab file has been created.
        my $keytab = GetKeytabName($hosts, $principal_name, $keytab_file, $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR});

        my $command = "test -f $keytab";

        if ($OPT_DRY_RUN)
        {
            print "$command\n";
            next;
        }

        USER_INFO("Check if Kerberos keytab $keytab has been created...");
        my ($rc, $command_output) = RunCommand($command);

        VERBOSE_INFO($command_output) if ($command_output);
        if ($rc)
        {
            my $msg = "Failed to find Kerberos keytab: $rc";
            USER_ERROR($msg);
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    return 0;
}

# Notes:
# - It is possible to customize the path of each keytab file on Ambari. So the
#   keytabs file can be created outside /etc/secutiry/keytabs. Therefore, we
#   transfer keytab files one by one.
# - Customers can modify this script to use other file transfer programs.
# - For simplicity, the same command is used for both local and remote hosts.
sub TransferKeytabs($)
{
    my $services = $_[0];

    my $unix_user = "root";

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $principal_name = $service->{'principal_name'};
        my $keytab_file = $service->{'keytab_file'};
        my $keytab_dir = $service->{'keytab_dir'};

        my $keytab = GetKeytabName($hosts, $principal_name, $keytab_file, $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR});
        my $keytab_dest = File::Spec->catpath("", $keytab_dir, $keytab_file);

        foreach my $host (sort keys %$hosts)
        {
            my $command = GetSCPCommand($host, $keytab, $keytab_dest);

            if ($OPT_DRY_RUN)
            {
                print "$command\n";
                next;
            }

            USER_INFO("Transferring Kerberos keytab $keytab to host $host...");
            my ($rc, $command_output) = RunCommand($command);

            VERBOSE_INFO($command_output) if ($command_output);
            if ($rc)
            {
                my $msg = "Failed to transfer Kerberos keytab: $rc";
                USER_ERROR($msg);
                return $rc unless $OPT_IGNORE_ERROR;
            }
        }
    }

    return 0;
}

sub ChownKeytabs($)
{
    my $services = $_[0];

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $keytab_file = $service->{'keytab_file'};
        my $keytab_dir = $service->{'keytab_dir'};
        my $user = $service->{'user'};
        my $group = $service->{'group'};

        my $keytab_dest = File::Spec->catpath("", $keytab_dir, $keytab_file);

        foreach my $host (sort keys %$hosts)
        {
            my $command = GetSSHCommand($host, "chown $user:$group $keytab_dest");

            if ($OPT_DRY_RUN)
            {
                print "$command\n";
                next;
            }

            USER_INFO("Changing ownership of Kerberos keytab $keytab_dest to $user:$group...");
            my ($rc, $command_output) = RunCommand($command);

            VERBOSE_INFO($command_output) if ($command_output);
            if ($rc)
            {
                my $msg = "Failed to change ownership of Kerberos keytab: $rc";
                USER_ERROR($msg);
                return $rc unless $OPT_IGNORE_ERROR;
            }
        }
    }

    return 0;
}

sub ChmodKeytabs($)
{
    my $services = $_[0];

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $keytab_file = $service->{'keytab_file'};
        my $keytab_dir = $service->{'keytab_dir'};
        my $user = $service->{'user'};
        my $group = $service->{'group'};
        my $perm = $service->{'perm'};

        my $keytab_dest = File::Spec->catpath("", $keytab_dir, $keytab_file);

        foreach my $host (sort keys %$hosts)
        {
            my $command = GetSSHCommand($host, "chmod $perm $keytab_dest");

            if ($OPT_DRY_RUN)
            {
                print "$command\n";
                next;
            }

            USER_INFO("Changing permission of Kerberos keytab $keytab_dest to $perm...");
            my ($rc, $command_output) = RunCommand($command);

            VERBOSE_INFO($command_output) if ($command_output);
            if ($rc)
            {
                my $msg = "Failed to change permission of Kerberos keytab: $rc";
                USER_ERROR($msg);
                return $rc unless $OPT_IGNORE_ERROR;
            }
        }
    }

    return 0;
}

# Check to ensure CDC is ready on all cluster nodes.
sub CheckClusterCDCStatus($)
{
    my $cluster = $_[0];
    my $rc = 0;

    foreach my $host (sort keys %$cluster)
    {
        $rc = CheckCDCStatus(1, $host);
        return $rc if ($rc);
    }

    # Return success if no host in the cluster
    return $rc;
}

# Check to ensure CDC is ready locally.
sub CheckLocalCDCStatus()
{
    # Note that we pass "localhost" to CheckCDCStatus() solely for logging
    return CheckCDCStatus(0, "localhost");
}

#
# Check if CDC is up and connected to AD.
#
#   $_[0]: $is_remote   - Whether to run remote command (using SSH).
#   $_[1]: $host        - Hostname to run remote command. Used only when
#                         $is_remote is true.
#
#   return: 0           - Success
#           <errno>     - Failure
#
sub CheckCDCStatus($$)
{
    my $is_remote = $_[0];
    my $host = $_[1];

    #
    # adinfo -m will return errno as follows:
    # 0: connected
    # 1: disconnected
    # ...
    # 7: <unavailable> (e.g. adclient stopped)
    #
    my $command = "$CIMS_USR/bin/adinfo -m";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        return 0;
    }

    USER_INFO("Checking CentrifyDC status on host $host...");
    my ($rc, $command_output) = RunCommand($command);

    VERBOSE_INFO($command_output) if ($command_output);
    if ($rc)
    {
        my $msg = "CentrifyDC is NOT connected to AD: $rc";
        USER_ERROR($msg);
        return $rc unless $OPT_IGNORE_ERROR;
    }

    return 0;
}

# Delete specified SPN(s) from AD computer object on all cluster nodes.
# Note that may need to delete nfs as well depending on hadoop.conf option
sub DeleteSPN($)
{
    my $remove_ftp = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_FTP_REMOVE};
    my $remove_cifs = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_CIFS_REMOVE};
    my $remove_nfs = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_NFS_REMOVE};
    my $remove_http = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_HTTP_REMOVE};

    if (!$remove_ftp && !$remove_cifs && !$remove_nfs && !$remove_http)
    {
        VERBOSE_DEBUG("None of configuration parameters set to remove SPN - Do nothing.");
        return 0;
    }

    my $cluster = $_[0];

    foreach my $host (sort keys %$cluster)
    {
        # Using machine credential so will not ask for password
        my $command = "$ADKEYTAB -m --delspn";
        $command = "$command -P ftp" if ($remove_ftp);
        $command = "$command -P cifs" if ($remove_cifs);
        $command = "$command -P nfs" if ($remove_nfs);
        $command = "$command -P http" if ($remove_http);
        $command = GetSSHCommand($host, $command);

        if ($OPT_DRY_RUN)
        {
            print "$command\n";
            next;
        }

        USER_INFO("Deleting SPN(s) from AD computer object on host $host...");
        my ($rc, $command_output) = RunCommand($command);

        VERBOSE_INFO($command_output) if ($command_output);
        if ($rc)
        {
            my $msg = "Failed to delete SPN: $rc";
            USER_ERROR($msg);
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    return 0;
}

#
# Update centrifydc.conf on all cluster nodes.
#
# If parameter value "http" is found from adclient.krb5.service.principals,
#   Will remove the value "http".
# Else the following settings will be added to config file:
# adclient.krb5.service.principals: ftp cifs nfs
#
# Note that nfs may or may not be there depending on the option in hadoop.conf
# Note that corresponding GPs should NOT be enabled if this function is run.
#
# TODO: Allow end users to choose whether to edit centrifydc.conf by this
# script.
#
sub ModifyAdclientConfig($)
{
    my $remove_ftp = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_FTP_REMOVE};
    my $remove_cifs = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_CIFS_REMOVE};
    my $remove_nfs = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_NFS_REMOVE};
    my $remove_http = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_ADCLIENT_KRB5_SERVICE_PRINCIPAL_HTTP_REMOVE};

    if (!$remove_ftp && !$remove_cifs && !$remove_nfs && !$remove_http)
    {
        VERBOSE_DEBUG("None of configuration parameters set to remove SPN - Do nothing.");
        return 0;
    }

    my $cluster = $_[0];
 
    # CLI adproperty will handle the write lock while writing centrifydc.conf
    foreach my $host (sort keys %$cluster)
    {
        # Try to get the valid parameter setting. 
        my $command_get_property = "$ADPROPERTY --get $PROP_CDC_KEY_KRB5_SERVICE_PRINCIPALS";
        # Add an extra line to separate SSH banner info from SSH command output.
        $command_get_property = "echo $SSH_BANNER_PATTERN; " .  $command_get_property; 
        $command_get_property = GetSSHCommand($host, $command_get_property);

        # Try to set expected parameter values.
        # If setting is found, filter out 'http' if it contains "http".
        # If setting is not found, set default value as "ftp cifs nfs" in centrifydc.conf
        # Note that nfs may be treated same as http depending on the option in hadoop.conf
        my $adproperty_set_string = "$ADPROPERTY --set $PROP_CDC_KEY_KRB5_SERVICE_PRINCIPALS %s";
        $adproperty_set_string = GetSSHCommand($host, $adproperty_set_string);

        # Use default if cdc.conf property is not active.
        # Note that $def_setting should exclude nfs too depending on the option in hadoop.conf

        my $def_setting = "";
        $def_setting = "$def_setting --data ftp" if (!$remove_ftp);
        $def_setting = "$def_setting --data cifs" if (!$remove_cifs);
        $def_setting = "$def_setting --data nfs" if (!$remove_nfs);
        $def_setting = "$def_setting --data http" if (!$remove_http);

        my $command_set_property = sprintf($adproperty_set_string, $def_setting);       
        if ($OPT_DRY_RUN)
        {
            print "$command_get_property\n";
            print "$command_set_property\n";
            next;
        }

        USER_INFO("Getting property $PROP_CDC_KEY_KRB5_SERVICE_PRINCIPALS from host $host...");
        my ($rc_modify, $command_output) = RunCommand($command_get_property);
        VERBOSE_INFO($command_output) if ($command_output);
        # $rc_modify - 0     Command success, property is found.
        #            - 156   Command success, property is NOT found.
        #            - other Unknown command failure.
        if ($rc_modify && $rc_modify != $ERR_ADPROP_NOT_FOUND)
        {
            my $msg = "Failed to get property: $rc_modify";
            USER_ERROR($msg);
            return $rc_modify if (!$OPT_IGNORE_ERROR);
        }
        if (!$rc_modify)
        {
            my $new_values = "";
            my $prop_value_verbose = "";
            my $escape_ssh_banner = 0;
            my $spn_found = 0;
            foreach (split("\n", $command_output))
            {
                s/\r//g;
                # We need to escape SSH banner info, since they're not wanted
                if (!$escape_ssh_banner)
                {
                    $escape_ssh_banner = 1 if (m/^$SSH_BANNER_PATTERN$/);
                    next;
                }
                # skip any SPN(s) that is configured to be removed
                if (m/\bftp\b/)
                {
                    if ($remove_ftp)
                    {
                        $spn_found = 1;
                        next;
                    }
                }
                if (m/\bcifs\b/)
                {
                    if ($remove_cifs)
                    {
                        $spn_found = 1;
                        next;
                    }
                }
                if (m/\bnfs\b/)
                {
                    if ($remove_nfs)
                    {
                        $spn_found = 1;
                        next;
                    }
                }
                if (m/\bhttp\b/)
                {
                    if ($remove_http)
                    {
                        $spn_found = 1;
                        next;
                    }
                }
                # Meet the tradeoff line. We already get all command result now.
                last if (m/Connection to \S+ closed./);
                $new_values .= " --data \\\"$_\\\"";
                $prop_value_verbose .= " $_";
            }
            # This should not occur since SSH banner separator should be echo-ed
            if (!$escape_ssh_banner)
            {
                WARN_OUT("Failed to get SSH banner separator from command output.");
                next;
            }
            if (!$spn_found)
            {
                # If parameter values are expected, do nothing.
                $prop_value_verbose =~ s/^\s+//;  # Trim the white space
                VERBOSE_DEBUG("Property $PROP_CDC_KEY_KRB5_SERVICE_PRINCIPALS is found with expected value: [$prop_value_verbose].");
                next;
            }
            $command_set_property = sprintf($adproperty_set_string, $new_values); 
        }

        # Update appropriate <para: value> to cdc.conf.
        USER_INFO("Setting property $PROP_CDC_KEY_KRB5_SERVICE_PRINCIPALS to host $host...");
        ($rc_modify, $command_output) = RunCommand($command_set_property);
        # $rc_modify - 0     Command success, property is set.
        #            - 9     Root privilege required.
        #            - 157   Fail to set property.
        #            - other Unknown command failure.
        VERBOSE_INFO($command_output) if ($command_output);
        if ($rc_modify)
        {
            my $msg = "Failed to set property: $rc_modify";
            USER_ERROR($msg);
            return $rc_modify if (!$OPT_IGNORE_ERROR);
        }
    }

    return 0;
}

sub CreateLocalKeytabDir($)
{
    my $dir = $_[0];

    my $rc;

    $rc = CheckKeytabDirForCreation(0, "localhost", $dir);
    # Skip creation if directory already exists (not in dry-run mode)
    return 0 if ($rc);

    $rc = CreateKeytabDir(0, "localhost", $dir);
    return $rc if ($rc);

    # Since this directory is for local data, there is no need to chown to
    # hadoop group as in CreateClusterKeytabDirs(). 

    $rc = ChmodKeytabDir(0, "localhost", $dir);
    return $rc if ($rc);

    return 0;
}

#
# Create data directory (and its sub-directories) locally before deploying
# keytabs to all cluster nodes. These directories will be kept until keytab
# files (and their AD objects) are deleted using this script.
#
sub CreateLocalKeytabDirs($)
{
    my $cluster = $_[0];
  
    if ($OPT_DRY_RUN)
    {
        VERBOSE_DEBUG("VERBOSE: Assume local data directories do not exist in dry-run mode.");
    }

    #
    # Create local data directories
    #

    my $rc = 0;

    # Data directory
    $rc = CreateLocalKeytabDir($CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_DIR});
    if ($rc)
    {
        return $rc unless $OPT_IGNORE_ERROR;
    }

    # Data sub-directory
    $rc = CreateLocalKeytabDir($CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR});
    if ($rc)
    {
        return $rc unless $OPT_IGNORE_ERROR;
    }

    # Per-host directories
    foreach my $host (sort keys %$cluster)
    {
        my $short_hostname = GetShortHostName($host, undef);
        my $local_dir = "$CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR}/$short_hostname";

        $rc = CreateLocalKeytabDir($local_dir);
        if ($rc)
        {
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    return 0;
}

#
# Ensure keytab directories exist and with the right ownership/permission on
# all cluster nodes.
#
# Note:
#
# - Hortonworks uses /etc/security/keytabs/ by default. But can be modified
#   when security is enabled on Ambari.
#
# - See below for the recommended ownership/permission:
#
#   mkdir -p /etc/security/keytabs/
#   chown root:hadoop /etc/security/keytabs
#   chmod 750 /etc/security/keytabs
#
#   http://ambari.apache.org/current/installing-hadoop-using-ambari/content/ambari-kerb-1-4.html
#
# - Therefore, this function only ensures the direct parent directories has the
#   right ownership/permission. For example, if all keytabs files are
#   configured to /etc/security/keytabs/hadoop/ from CSV file. Only the
#   "hadoop" folder will have ownership/permission configured by this function.
#
sub CreateClusterKeytabDirs($)
{
    my $cluster = $_[0];

    if ($OPT_DRY_RUN)
    {
        # For "test -d"
        VERBOSE_DEBUG("VERBOSE: Assume Kerberos keytab directory does not exist on all hosts in dry-run mode.");
    }

    #
    # Create keytab directory on each cluster node
    #

    foreach my $host (sort keys %$cluster)
    {
        my $node = $cluster->{$host};

        my $keytab_dirs = $node->{'keytab_dirs'};
        my $group = $node->{'group'};

        my $rc;

        foreach my $keytab_dir (sort keys %$keytab_dirs)
        {
            # For simplicity, we always run these commands without checking
            # current status.
            
            $rc = CheckKeytabDirForCreation(1, $host, $keytab_dir);
            # Skip creation if directory already exists (not in dry-run mode)
            next if ($rc);
            
            $rc = CreateKeytabDir(1, $host, $keytab_dir);
            return $rc if ($rc);

            $rc = ChownKeytabDir(1, $host, $keytab_dir, $group);
            return $rc if ($rc);

            $rc = ChmodKeytabDir(1, $host, $keytab_dir);
            return $rc if ($rc);
        }
    }

    return 0;
}

#
# Test if destination path is a directory.
#
#   $_[0]:  $is_remote - 0: local host;
#                        1: remote host.
#   $_[1]:  $host - the remote host name.
#   $_[2]:  $path - absolute path of the remote dir.
#
#   return - 1:     path is a dir.
#          - 0:     path is not a dir.
#
sub CheckKeytabDirForDeletion($$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $keytab_dir = $_[2];

    my $command = "test -d $keytab_dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        # Assume that dir always exists in dry run mode.
        # So we will delete it later.
        return 1;
    }

    USER_INFO("Checking if directory $keytab_dir exists on host $host...");
    my ($rc, $command_output) = RunCommand($command);
    if ($rc == 0)
    {
        return 1;
    }

    USER_INFO("Directory $keytab_dir does not exist on host $host...");
    return 0;
}

#
# Test if destination directory exists or not.
#
#   $_[0]:  $is_remote - 0: local host;
#                        1: remote host.
#   $_[1]:  $host - the remote host name.
#   $_[2]:  $path - absolute path of the remote dir.
#
#   return - 1:     dir already exists.
#          - 0:     dir not exists.
#
sub CheckKeytabDirForCreation($$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $keytab_dir = $_[2];

    my $command = "test -d $keytab_dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        # Assume that dir always NOT exists in dry run mode.
        # So we will create it later.
        return 0;
    }

    USER_INFO("Checking if directory $keytab_dir exists on host $host...");
    my ($rc, $command_output) = RunCommand($command);
    if ($rc == 0)
    {
        USER_INFO("Directory $keytab_dir already exists on host $host...");
        return 1;
    }

    return 0;
}

sub CreateKeytabDir($$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $keytab_dir = $_[2];

    my $command = "mkdir -p $keytab_dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        return 0;
    }

    USER_INFO("Creating directory $keytab_dir on host $host...");
    my ($rc, $command_output) = RunCommand($command);

    VERBOSE_INFO($command_output) if ($command_output);
    if ($rc)
    {
        my $msg = "Failed to create directory: $rc";
        USER_ERROR($msg);
        return $rc unless $OPT_IGNORE_ERROR;
    }

    return 0;
}

sub ChownKeytabDir($$$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $keytab_dir = $_[2];
    my $group = $_[3];

    my $command = "chown root:$group $keytab_dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        return 0;
    }

    USER_INFO("Changing ownership of directory $keytab_dir to root:$group...");
    my ($rc, $command_output) = RunCommand($command);

    VERBOSE_INFO($command_output) if ($command_output);
    if ($rc)
    {
        my $msg = "Failed to change ownership of directory: $rc";
        USER_ERROR($msg);
        return $rc unless $OPT_IGNORE_ERROR;
    }

    return 0;
}

sub ChmodKeytabDir($$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $keytab_dir = $_[2];

    my $command = "chmod 750 $keytab_dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        return 0;
    }

    USER_INFO("Changing permission of directory $keytab_dir to 750...");
    my ($rc, $command_output) = RunCommand($command);

    VERBOSE_INFO($command_output) if ($command_output);
    if ($rc)
    {
        my $msg = "Failed to change ownership of directory: $rc";
        USER_ERROR($msg);
        return $rc unless $OPT_IGNORE_ERROR;
    }

    return 0;
}

sub DeleteKeytabDir($$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $keytab_dir = $_[2];

    # For simplicity, we always run this command without checking if e.g.
    # directory exists.
    my $command = "rmdir";
    $command = "$command $keytab_dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        return 0;
    }

    USER_INFO("Deleting directory $keytab_dir on host $host...");
    my ($rc, $command_output) = RunCommand($command);

    VERBOSE_INFO($command_output) if ($command_output);
    if ($rc)
    {
        my $msg = "Failed to delete directory: $rc";

        # Note that we simply try to remove the directory if empty. So we just
        # leave the directory there if not empty. Be subtle and log in INFO
        # level.
        USER_INFO($msg);
    }

    # NOT an error if failed to remove directory
    return 0;
}

sub DeleteClusterKeytabs($)
{
    my $services = $_[0];

    foreach my $key (sort keys %$services)
    {
        my $service = $services->{$key};

        my $hosts = $service->{'hosts'};
        my $keytab_file = $service->{'keytab_file'};
        my $keytab_dir = $service->{'keytab_dir'};

        my $keytab_dest = File::Spec->catpath("", $keytab_dir, $keytab_file);

        foreach my $host (sort keys %$hosts)
        {
            my $command = "rm";
            $command = "$command --force" if $OPT_FORCE;
            $command = "$command $keytab_dest";
            $command = GetSSHCommand($host, $command);

            if ($OPT_DRY_RUN)
            {
                print "$command\n";
                next;
            }

            USER_INFO("Deleting Kerberos keytab $keytab_dest from host $host...");
            my ($rc, $command_output) = RunCommand($command);

            VERBOSE_INFO($command_output) if ($command_output);
            if ($rc)
            {
                my $msg = "Failed to delete Kerberos keytab: $rc";
                USER_ERROR($msg);
                return $rc unless $OPT_IGNORE_ERROR;
            }
        }
    }

    return 0;
}

#
# Fetch the pure ran command result from SSH command output.
# e.g. filter out the trade off messages like:
#  Connection to $host closed.
#
#   $_[0]: - $output: original SSH command output.
#
#   return - $result: pure ran command result.
#
sub PostProcessSSHCommandOutput($)
{
    my $output = $_[0];
    my $result = "";
    foreach (split("\n", $output))
    {
        s/\r//g;
        # Meet the tradeoff line. We already get all command result now.
        last if (m/^Connection to \S+ closed.$/);
        $result .= "$_\n";
    }
    chomp $result;

    return $result;
}


#
# Test if remote dir is an empty dir.
# Please make sure the passed in $path is a dir.
#
#   $_[0]:  $is_remote - 0: local host;
#                        1: remote host.
#   $_[1]:  $host - the remote host name.
#   $_[2]:  $path - absolute path of the remote dir.
#
#   return - 1:     path is an empty dir.
#          - 0:     path is not empty.
#          - undef: undefined error.
#
sub IsKeytabDirEmpty($$$)
{
    my $is_remote = $_[0];
    my $host = $_[1];
    my $dir = $_[2];

    # Use command "ls -A" to test if directory is empty.
    # So we can list all directory contents but the implied . and ..
    # Since command "ls -l" will always shows pattern "total \d+" as the first
    # line, we can use this pattern to locate the expected line position of
    # command output. This will also help to filter out the noise messages from
    # SSH banner.
    my $command = "ls -lA $dir";
    $command = GetSSHCommand($host, $command) if ($is_remote);

    if ($OPT_DRY_RUN)
    {
        print "$command\n";
        # Assume that dir is empty in dry run mode.
        # So we will delete it later.
        return 1;
    }

    my ($rc, $command_output) = RunCommand($command);
    VERBOSE_INFO($command_output) if ($command_output);
    $command_output = PostProcessSSHCommandOutput($command_output) if ($is_remote);
    if ($rc)
    {
        USER_WARN("Failed to list directory $dir on host $host");
        return undef;
    }

    my $is_pattern_found = 0;
    my $line_num = 0;
    foreach (split("\n", $command_output))
    {
        # For remote command output, we will step over the noise messages from
        # SSH banner. Since the banner will always be in front of pattern
        # "total \d+"
        if (/^total \d+$/)
        {
            $is_pattern_found = 1;
            next;
        }
        next if (!$is_pattern_found);
        $line_num++;
        # No need to count all line numbers. Just need to confirm the specified
        # dir is not empty.
        last;
    }
    if ($line_num)
    {
        VERBOSE_DEBUG("Directory $dir on host $host is not empty.");
        return 0;
    }
    return 1;
}

sub DeleteLocalKeytabDirs($)
{
    my $cluster = $_[0];

    #
    # Delete local keytab directories
    #

    my $rc = 0;

    # Per-host directories
    foreach my $host (sort keys %$cluster)
    {
        my $short_hostname = GetShortHostName($host, undef);
        my $local_dir = "$CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR}/$short_hostname";

        # Make sure that it is a directory.
        next if (!CheckKeytabDirForDeletion(0, "localhost", $local_dir));
        # Do not try to remove the directory if not empty. 
        next if (!IsKeytabDirEmpty(0, "localhost", $local_dir));

        $rc = DeleteKeytabDir(0, "localhost", $local_dir);
        if ($rc)
        {
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    my $local_data_sub_dir = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_SUB_DIR};
    # Remove the directory if empty. 
    if (CheckKeytabDirForDeletion(0, "localhost", $local_data_sub_dir) && IsKeytabDirEmpty(0, "localhost", $local_data_sub_dir))
    {
        # Data sub-directory
        $rc = DeleteKeytabDir(0, "localhost", $local_data_sub_dir);
        if ($rc)
        {
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    my $local_data_dir = $CentrifyDC::Hadoop::Config::g_props->{PROP_KEY_DATA_DIR};
    # Remove the directory if empty. 
    if (CheckKeytabDirForDeletion(0, "localhost", $local_data_dir) && IsKeytabDirEmpty(0, "localhost", $local_data_dir))
    {
        # Data directory
        $rc = DeleteKeytabDir(0, "localhost", $local_data_dir);
        if ($rc)
        {
            return $rc unless $OPT_IGNORE_ERROR;
        }
    }

    return 0;
}

sub DeleteClusterKeytabDirs($)
{
    my $cluster = $_[0];

    #
    # Delete keytab directory on each cluster node if empty
    #

    foreach my $host (sort keys %$cluster)
    {
        my $node = $cluster->{$host};
        my $keytab_dirs = $node->{'keytab_dirs'};

        foreach my $keytab_dir (sort keys %$keytab_dirs)
        {
            # Make sure that it is a directory.
            next if (!CheckKeytabDirForDeletion(1, $host, $keytab_dir));
            # Do not try to remove the directory if not empty. 
            next if (!IsKeytabDirEmpty(1, $host, $keytab_dir));
            my $rc = DeleteKeytabDir(1, $host, $keytab_dir);
            return $rc if $rc;
        }
    }

    return 0;
}

#
# Check if root privilege is provided.
#
# Note that we do not check this in dry run. So all users can generate and
# review the commands.
#
sub CheckRootPrivilege()
{
    if ($OPT_DRY_RUN)
    {
        VERBOSE_DEBUG("VERBOSE: Skip root privilege checking in dry-run mode.");
        return 0;
    }

    if ($> != 0)
    {
        my $msg = "ERROR: Root privilege required.";
        USER_ERROR($msg);
        return 1 unless $OPT_IGNORE_ERROR;
    }

    return 0;
}

#
# Check if valid TGT exists in cred cache. Since adkeytab will ask for
# domain admin privilege if no valid TGT exists.
#
# If valid TGT can be found, run klist to show the default principal currently
# using.
#
# Note that unique ccache files are NOT supported, since Hadoop services are
# not using them.
#
sub CheckValidTGT()
{
    my $command1 = "$CENTRIFY_KRB5_BIN_PATH/klist -s";
    my $command2 = "$CENTRIFY_KRB5_BIN_PATH/klist";

    if ($OPT_DRY_RUN)
    {
        print "$command1\n";
        print "$command2\n";
        return 0;
    }

    my ($rc, $command_output) = RunCommand($command1);

    VERBOSE_INFO($command_output) if ($command_output);
    if ($rc)
    {
        my $msg = "Failed to find any valid TGT in Kerberos credential cache: $rc";
        USER_ERROR($msg);
        return $rc unless $OPT_IGNORE_ERROR;
    }

    # Even if the TGT is valid, the principal might be wrong
    # (e.g.HTTP/host1.example.com@EXAMPLE.COM). So show the default principal
    # for debugging purpose.
    
    ($rc, $command_output) = RunCommand($command2);

    if ($rc)
    {
        # klist will return 1 when no credentials cache found. But klist -s
        # should return 1 already. So such case will NOT appear here.
        VERBOSE_INFO($command_output) if ($command_output);

        my $msg = "Failed to list Kerberos credential cache: $rc";
        USER_ERROR($msg);
        return $rc unless $OPT_IGNORE_ERROR;
    }

    $command_output =~ /Default principal:\s+(\S+)/;
    my $default_principal = $1;

    if (!defined($default_principal))
    {
        VERBOSE_INFO($command_output) if ($command_output);

        my $msg = "Failed to get default principal from Kerberos credential cache";
        USER_ERROR($msg);
        return 1 unless $OPT_IGNORE_ERROR;
    }

    # We don't really care if the ccache name can be found or not. Just for
    # reporting.
    $command_output =~ /Ticket cache:\s+(\S+)/;
    my $ticket_cache = $1;

    USER_INFO("Default principal $default_principal from ticket cache $ticket_cache will be used.");

    return 0;
}

1;
