#!/bin/bash

# RHEL/CentOS executes authconfig to config smart card.
# ex:
#   enable smart card
#       authconfig --enablesmartcard   --update --nostart
#   disable smart card
#       authconfig --disablesmartcard  --update --nostart
#   smart card removal action: Ignore
#       authconfig --smartcardaction=0 --update --nostart
#   smart card removal action: Lock
#       authconfig --smartcardaction=1 --update --nostart
#
# But authconfig may changes some system config files.
# ex: authconfig re-create /etc/nsswitch.conf without centrify patches.
# This causes RHEL cannot login with AD user or smart card.
#
# This script skips authconfig command and configs smart card manually. 

GREP="/bin/grep"
MKDIR="/bin/mkdir"
MV="/bin/mv"
RM="/bin/rm"
RPM="/bin/rpm"
SED="/bin/sed"

usage() {
    echo "usage: `basename $0` <config_name> <config_value>"
    echo "  config_name     | config_value    "
    echo "  --------------- + --------------  "
    echo "  smartcard       | enable|disable  "
    echo "  smartcardaction | ignore|lock     "
    exit 1
}

if [ ! -f /etc/redhat-release ]; then
    exit 0
fi

# RHEL/CentOS major version, and smart card supported on 5 and above
RHEL_RPM_PACKAGE=`$RPM -q --whatprovides /etc/redhat-release`
if [ ! -n "$RHEL_RPM_PACKAGE" ]; then
    exit 0
fi
RHEL_MAJOR_VERSION=`$RPM -q --queryformat '%{VERSION}' $RHEL_RPM_PACKAGE | $SED -e 's#^\([0-9]*\).*#\1#'`
if [ ! -n "$RHEL_MAJOR_VERSION" ]; then
    exit 0
fi
if [ $RHEL_MAJOR_VERSION -lt 5 ]; then
    exit 0
fi

if [ $# -ne 2 ]; then
    usage
fi

CONFIG_NAME=$1
CONFIG_VALUE=$2

ARGS_IS_VALID=0
if [ "$CONFIG_NAME" == "smartcard" ]; then
    if [ "$CONFIG_VALUE" == "enable" ]; then
        ARGS_IS_VALID=1
    elif [ "$CONFIG_VALUE" == "disable" ]; then
        ARGS_IS_VALID=1
    fi
elif [ "$CONFIG_NAME" == "smartcardaction" ]; then
    if [ "$CONFIG_VALUE" == "ignore" ]; then
        ARGS_IS_VALID=1
    elif [ "$CONFIG_VALUE" == "lock" ]; then
        ARGS_IS_VALID=1
    fi
fi

if [ $ARGS_IS_VALID -ne 1 ]; then
    usage
fi

AUTHCONFIG_CONF="/etc/sysconfig/authconfig"
PKCS11_SETUP_CMD="/usr/bin/pkcs11_setup"
DCONF_CMD="/usr/bin/dconf"
GCONFTOOL_CMD="/usr/bin/gconftool-2"

if [ ! -f $AUTHCONFIG_CONF ]; then
    exit 1
fi
if [ ! -x $PKCS11_SETUP_CMD ]; then
    exit 1
fi
if [ ! -x $GCONFTOOL_CMD ]; then
    exit 1
fi

writeDConf() {
    if [ $RHEL_MAJOR_VERSION -lt 7 ]; then
        return
    fi

    DCONF_DIR="/etc/dconf/db/distro.d"
    DCONF_CONF=$DCONF_DIR"/10-authconfig"
    DCONF_LOCKS_DIR=$DCONF_DIR"/locks"
    DCONF_LOCKS_CONF=$DCONF_LOCKS_DIR"/10-authconfig-locks"

    if [ ! -d $DCONF_DIR ]; then
        $MKDIR -p $DCONF_DIR
    fi
    if [ ! -d $DCONF_LOCKS_DIR ]; then
        $MKDIR -p $DCONF_LOCKS_DIR
    fi
    if [ -f $DCONF_CONF ]; then
       $RM -f $DCONF_CONF 
    fi
    if [ -f $DCONF_LOCKS_CONF ]; then
       $RM -f $DCONF_LOCKS_CONF 
    fi

    need_config=0
    enableSmartcard=0
    forceSmartcard=0
    enableFprintd=0

    if [ "`$GREP '^USESMARTCARD=yes' $AUTHCONFIG_CONF`""no" != "no" ]; then
        enableSmartcard=1
    else
        need_config=1
    fi
    if [ "`$GREP '^FORCESMARTCARD=yes' $AUTHCONFIG_CONF`""no" != "no" ]; then
        forceSmartcard=1
        need_config=1
    fi
    if [ "`$GREP '^USEFPRINTD=yes' $AUTHCONFIG_CONF`""no" != "no" ]; then
        enableFprintd=1
    else
        need_config=1
    fi

    if [ $need_config -eq 1 ]; then
        echo "# Generated by $0" >> $DCONF_CONF
        echo "" >> $DCONF_CONF
        echo "[org/gnome/login-screen]" >> $DCONF_CONF

        if [ $enableSmartcard -eq 0 ]; then
            echo "enable-smartcard-authentication=false" >> $DCONF_CONF
            echo "/org/gnome/login-screen/enable-smartcard-authentication" >> $DCONF_LOCKS_CONF
        fi

        if [ $forceSmartcard -eq 1 ]; then
            echo "enable-password-authentication=false" >> $DCONF_CONF
            echo "/org/gnome/login-screen/enable-password-authentication" >> $DCONF_LOCKS_CONF
        fi

        if [ $enableFprintd -eq 0 ]; then
            echo "enable-fingerprint-authentication=false" >> $DCONF_CONF
            echo "/org/gnome/login-screen/enable-fingerprint-authentication" >> $DCONF_LOCKS_CONF
        fi
    fi

    $DCONF_CMD update
}

case "$CONFIG_NAME" in
"smartcard")
    enable="no"
    if [ "$CONFIG_VALUE" == "enable" ]; then
        enable="yes"
    fi
    $SED -e 's/^USESMARTCARD=.*/USESMARTCARD='"$enable"'/g' $AUTHCONFIG_CONF > $AUTHCONFIG_CONF.cdc
    $MV -f $AUTHCONFIG_CONF.cdc $AUTHCONFIG_CONF

    writeDConf
    ;;
"smartcardaction")
    ins_action="/usr/sbin/gdm-safe-restart"
    rm_action="/usr/sbin/gdm-safe-restart"
    lock="none"

    if [ "$CONFIG_VALUE" == "lock" ]; then
        ins_action="${ins_action},/etc/pkcs11/lockhelper.sh -lock"
        rm_action="${rm_action},/etc/pkcs11/lockhelper.sh -deactivate"
        lock="lock_screen"
    fi

    $PKCS11_SETUP_CMD ins_action="$ins_action" rm_action="$rm_action"
    $GCONFTOOL_CMD  --direct \
                    --config-source=xml:readwrite:/etc/gconf/gconf.xml.mandatory \
                    -s /desktop/gnome/peripherals/smartcard/removal_action $lock \
                    --type string
    ;;
esac

exit 0
