#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2018 Centrify Corporation. All rights reserved.
#
# This is a common function used by other /cli/configurefiles/* scripts,
# which edit PAM files.
# It helps other scripts to load settings and configure files about PAM,
# and supplies all options and settings we need.
# It loads settings in three ways
# 1, by command line input
# 2, load setting from centrifydc
# 3, use DEFAULT value defined below.
#
##############################################################################

package autoedit_configure;
BEGIN{push(@INC, "/usr/share/centrifydc/scripts")};

use strict;
use warnings;
use Getopt::Long qw(GetOptions);

use autoedit_common_utils qw(loadSetting getValueList);
use autoedit_constants qw($CDC_Modified_Comment $CDC_Version);

use base 'Exporter';
our @EXPORT_OK = qw($cdc_long_comment $cdc_short_comment $cfg_dir $cdc_pam_lib 
        $libs_excluded $newLines $newLines2 $pam_d_ignored $pam_d_extra
        @pam_d_common_confs);

sub usage
{
    print <<END;
This command is used to add Centrifydc code into PAM file.
usage: 
    --cfg_dir               the path of pam dir, the default is "/etc"
    --cdc_pam_lib           the pam lib of Centrifydc
    --auth_libs_excluded    the auth libs which don't like 
                            our code to be above them
    --pam_d_ignored         the pam.d files we ignored
    --pam_d_extra           the pam.d files that we will check again.
                            Add centrifydc code, if it is not 
                            password-auth or common-auth style.
    --do_load_cdc_conf      load settings from centrifdc.conf
END
    exit 1;
}

use constant DEFAULT_CDC_PAM_LIB => "pam_centrifydc.so";
use constant DEFAULT_PAM_D_IGNORED => ['useradd', 'shadow'];
use constant DEFAULT_PAM_D_EXTRA => ['ftp', 'pure-ftpd', 'vsftpd', 'wu-ftpd', 'dzdo'];
# TODO: we also have a copy of DEFAULT_LIBS_EXCLUDED in centrifydc.conf,
# which names pam.config.preceding.authlibs. If you change the default value here, 
# please also update centrifydc.conf code.
use constant DEFAULT_LIBS_EXCLUDED => ['pam_allowroot', 'pam_rootok'
            , 'pam_rhosts_auth', 'pam_wheel', 'pam_listfile'
            , 'pam_mount', 'libmapr_pam'];
use constant DEFAULT_CFG_DIR => "/etc";

our $cfg_dir;
our $cdc_pam_lib;
our $cdc_long_comment = "$CDC_Modified_Comment $CDC_Version\n";
our $cdc_short_comment = "# added by Centrifydc ";
our $libs_excluded = {};
my $auth_libs_excluded;
our $do_load_cdc_conf;
our $pam_d_ignored;
our $pam_d_extra;

# Deal with the critical files, some of them must exist if the system
# is using the /etc/pam.d directory for PAM config files.
our @pam_d_common_confs = ( "password-auth",
                            "system-auth",
                            "common-auth",
                            "common-account",
                            "common-password",
                            "common-session"
                          );


GetOptions( "cfg_dir=s" => \$cfg_dir,
            "cdc_pam_lib=s" => \$cdc_pam_lib,
            "auth_libs_excluded=s" => \$auth_libs_excluded,
            "pam_d_ignored=s" => \$pam_d_ignored,
            "pam_d_extra=s" => \$pam_d_extra,
            "do_load_cdc_conf" => \$do_load_cdc_conf,
            ) or usage;

if (!defined $cfg_dir)
{
    $cfg_dir = DEFAULT_CFG_DIR;
}

if (!defined $cdc_pam_lib)
{
    $cdc_pam_lib = DEFAULT_CDC_PAM_LIB;
}

if (defined $auth_libs_excluded)
{
    $auth_libs_excluded = [split /[,\s]+/, $auth_libs_excluded];
}
else
{
    if ($do_load_cdc_conf)
    {
        $auth_libs_excluded = getValueList("pam.config.preceding.authlibs");
    }

    if (!defined $auth_libs_excluded)
    {
        $auth_libs_excluded = DEFAULT_LIBS_EXCLUDED;
    }
}
$libs_excluded->{"auth"} = $auth_libs_excluded;
$libs_excluded->{"account"} = $auth_libs_excluded;
$libs_excluded->{"password"} = $auth_libs_excluded;
$libs_excluded->{"session"} = $auth_libs_excluded;

my $unix_cred = "";
if (-e "/usr/lib/security/pam_unix_cred.so")
{
    $unix_cred = " unix_cred";
}

our ($newLines, $newLines2);

$newLines = ["auth\tsufficient\t$cdc_pam_lib$unix_cred\n",
             "auth\trequisite\t$cdc_pam_lib deny\n",
             "account\tsufficient\t$cdc_pam_lib\n",
             "account\trequisite\t$cdc_pam_lib deny\n",
             "session\trequired\t$cdc_pam_lib homedir\n",
             "password\tsufficient\t$cdc_pam_lib try_first_pass\n",
             "password\trequisite\t$cdc_pam_lib deny\n"
            ];

# for now, only auth module need to care those libs
# who don't like our code to be above them
$newLines2 = ["auth\tsufficient\t$cdc_pam_lib try_first_pass\n",
              "auth\trequisite\t$cdc_pam_lib deny\n"
             ];


if(defined $pam_d_ignored)
{
    $pam_d_ignored = [split /[,\s]+/, $pam_d_ignored];
}
else
{
    if($do_load_cdc_conf)
    {
        $pam_d_ignored = getValueList('pam.config.program.ignore');
    }

    if(!defined $pam_d_ignored)
    {
        $pam_d_ignored = DEFAULT_PAM_D_IGNORED;
    }
}
# process $pam_d_ignored, it should be full path
if (!(grep /$cfg_dir/, @$pam_d_ignored))
{
    grep s/.*/$cfg_dir\/pam.d\/$_/, @$pam_d_ignored;
}

if(defined $pam_d_extra)
{
    $pam_d_extra = [split /[,\s]+/, $pam_d_extra];
}
else
{
    if($do_load_cdc_conf)
    {
        $pam_d_extra = getValueList('pam.config.program.check');
    }

    if(!defined $pam_d_extra)
    {
        $pam_d_extra = DEFAULT_PAM_D_EXTRA;
    }
}
# process $pam_d_extra, it should be full path
if (!(grep /$cfg_dir/, @$pam_d_extra))
{
    grep s/.*/$cfg_dir\/pam.d\/$_/, @$pam_d_extra;
}

1;
