#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2018 Centrify Corporation. All rights reserved.
#
# This script is used to modify irs.conf
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts";
use CentrifyDC::Include;
use Getopt::Long qw(GetOptions);

use autoedit_common_utils qw(loadSetting getValueList);
use autoedit_constants qw($CDC_Modified_Comment $CDC_Version);
use autoedit_file_lock;
use CentrifyDC::GP::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify irs.conf
    usage:
    --cdc_tag           the name of centrifydc lib
    --cfg_file          the full path of irs.conf
    --nss_xmaps         the maps to be updated
    --do_load_cdc_conf  set it if want to load the above settings 
                        from centrifydc.conf
    --do_revert         set it if want to revert our changes
END
    exit 1;
}

use constant DEFAULT_CFG_FILE => "/etc/irs.conf";
use constant DEFAULT_NSS_XMAPS => [];
use constant DEFAULT_CDC_TAG => "nss_cdc";
use constant DEFAULT_HOSTS_NETWORKS_LOOKUP_ORDER => 
        ["dns continue", "nis continue", "local"];
use constant DEFAULT_NISMAP_LOOKUP_ORDER => 
        ["nis continue", "local"];

my $cfgName = DEFAULT_CFG_FILE;
my $cfgHandler;
my $newName;
my $newHandler;

my $cdc_tag = DEFAULT_CDC_TAG;
my $nss_xmaps;
my $do_load_cdc_conf;
my $do_revert;
my $cdc_long_comment = $CDC_Modified_Comment;

my @nssKeys = ( "hosts", "netgroup", "services", "protocols", "networks");
my %nss_reactions = ();

GetOptions( "cdc_tag=s" => \$cdc_tag,
            "cfg_file=s" => \$cfgName,
            "nss_xmaps=s" => \$nss_xmaps,
            "do_load_cdc_conf" => \$do_load_cdc_conf,
            "do_revert" => \$do_revert,
            ) or usage;

my %keysToCheck = ();

# need not to check all options when do revert
if (!$do_revert)
{
    if (! -e $cfgName)
    {
        INFO_OUT("Did not find $cfgName, create a new one.");
        CreateFile("$cfgName", {mode => 0600});
    }

    if (defined $nss_xmaps)
    {
        $nss_xmaps = [split /[,\s]+/, $nss_xmaps];
    }
    else
    {
        if ($do_load_cdc_conf)
        {
            $nss_xmaps = getValueList("adclient.autoedit.nss.xmaps");
        }

        if (!defined $nss_xmaps)
        {
            $nss_xmaps = DEFAULT_NSS_XMAPS;
        }
    }

    # get reaction for nss maps
    $nss_reactions{"hosts"} = DEFAULT_HOSTS_NETWORKS_LOOKUP_ORDER;
    $nss_reactions{"netgroup"} = DEFAULT_NISMAP_LOOKUP_ORDER;
    $nss_reactions{"services"} = DEFAULT_NISMAP_LOOKUP_ORDER;
    $nss_reactions{"protocols"} = DEFAULT_NISMAP_LOOKUP_ORDER;
    $nss_reactions{"networks"} = DEFAULT_HOSTS_NETWORKS_LOOKUP_ORDER;

    foreach my $key (@$nss_xmaps)
    {
        if (grep /^$key$/, @nssKeys)
        {
            $keysToCheck{$key} = 1;
        }
    }
}

my $flock = autoedit_file_lock::new("$cfgName");
if (!defined $flock)
{
    ERROR_OUT("cannot lock file $cfgName");
    exit 1;
}

$cfgName = TraverseSymLink($cfgName);
if (!defined $cfgName)
{
    ERROR_OUT("meet error when check symlink $cfgName");
    exit 1;
}

my $cfgLines = ReadFile("$cfgName");
if (!defined $cfgLines)
{
    ERROR_OUT("cannot read $cfgName");
    exit 1;
}

my @cfgLines = split "\n", $cfgLines;

if ($do_revert)
{
    @cfgLines = grep {$_ !~ /^$cdc_long_comment/} @cfgLines;
    @cfgLines = grep {$_ !~ /\b$cdc_tag\b/} @cfgLines;
}
else
{
    my @newLines = ();

    # Have any db been using $cdc_tag ?
    foreach my $key (keys %keysToCheck)
    {
        if (grep /^\s*$key\b.*\b$cdc_tag\b/, @cfgLines)
        {
            delete $keysToCheck{$key};
        }
    }
    if (scalar(keys %keysToCheck))
    {
        push @newLines, "$cdc_long_comment $CDC_Version";
    }
    else
    {
        # nothing need to do
        # all db are using $cdc_tag
        exit 0;
    }

    # update cfg file
    foreach my $key (keys %keysToCheck)
    {
        push @newLines, "$key $cdc_tag continue";
        if (grep /^\s*$key\b/, @cfgLines)
        {}
        else
        {
            my @reactions = @{$nss_reactions{$key}};
            push @newLines, (grep {$_ = "$key $_"} @reactions);
        }
    }

    unshift @cfgLines, @newLines;
}

$cfgLines = join "\n", @cfgLines;
$cfgLines .= "\n";

my $backup = "$cfgName.pre_cdc";
if (!$do_revert)
{
    if (! CopyFile($cfgName, $backup))
    {
        ERROR_OUT("Cannot backup $cfgName");
        exit 1;
    }
}
else
{
    # install.sh and rpm/pkg scripts may override config files with
    # the backup files if found them exist.
    # To avoid affecting these code logic, we'd better delete
    # backup file after "do_revert"
    if (-e $backup)
    {
        unlink $backup;
    }
}

if (!WriteFile($cfgName, $cfgLines))
{
    ERROR_OUT("Cannot write $cfgName");
    exit 1;
}

exit 0;

