#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2018 Centrify Corporation. All rights reserved.
#
# process pam.d files
# will go through all of the files under pam.d
#
# arguments:
#       please see autoedit_configure.pm
# return:
#       0 if succeed
#       1 if meet error
#
##############################################################################

BEGIN{push(@INC, "/usr/share/centrifydc/scripts")};

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use warnings;

use autoedit_pam_d qw(process_pam_d_file process_pam_d_files);
use autoedit_configure qw($cdc_long_comment $cfg_dir $cdc_pam_lib $libs_excluded 
                        $newLines $newLines2 $pam_d_ignored);

# process pam.d files
# will go through all of the files under pam.d
#
# $cdc_pam_lib      the name of Centrifydc pam lib
# $libs_excluded    the pam libs, which do not like our code to be above them
#                   It is a Ref of array
# $cfg_dir          the full path of pam.d directory
# $cdc_long_comment Centrifydc comment added before its code
# $newLines         our new code to add before $service_module
# $newLines2        our new code to add behind $service_module if found $libs_excluded
# $pam_d_ignored    the pam.d files we ignored
# return            undef if have anything wrong
#                   1 if succeed
#
sub process_pam_d_general
{
    my ($cdc_pam_lib, $libs_excluded, $cfg_dir, $cdc_long_comment, 
        $newLines, $newLines2, $pam_d_ignored) = @_;

    my $ret = 1;

    my @cfg_files = <$cfg_dir/*>;

    # these files need some extra keywords in pam settings
    # will deal with them seperately with other files
    my @exceptions = ("$cfg_dir/rsh",
                    "$cfg_dir/su",
                    "$cfg_dir/passwd"
                    );

    # skip *.pre_cdc backup file
    # skip @exceptions
    # skip $pam_d_ignored
    {
        my @files = ();
        foreach my $file (@cfg_files)
        {
            if ($file =~ /\.pre_cdc$/)
            {
                next;
            }
            if (grep /^$file$/, @exceptions)
            {
                next;
            }
            if (grep /^$file$/, @$pam_d_ignored)
            {
                next;
            }
            push @files, $file;
        }
        @cfg_files = @files;
    }

    # update "rsh" file
    # we expect rsh to have pam_rhosts_auth present as well, but it
    # does not need the "auth" whatsoever - rsh depends on
    # hosts.equiv and .rhosts. No other login check is done.
    my $cfg_file = "$cfg_dir/rsh";
    if (-e $cfg_file && !(grep /^$cfg_file$/, @$pam_d_ignored))
    {
        my @lines = @$newLines;
        @lines = grep {$_ !~ /\bauth\b/} @lines;
        my $rc = process_pam_d_file($cfg_file, \@lines, $libs_excluded, 
                    undef, $cdc_pam_lib, $cdc_long_comment, undef);
        if (!$rc)
        {
            # meet error
            $ret = $rc;
        }
    }

    # update "su" file
    # add keyword enable_dzpamgate
    $cfg_file = "$cfg_dir/su";
    if (-e $cfg_file && !(grep /^$cfg_file$/, @$pam_d_ignored))
    {
        my @lines = @$newLines;
        @lines = grep {$_ !~ /\bauth\b/} @lines;

        unshift @lines, (
            "auth\tsufficient\t$cdc_pam_lib enable_dzpamgate\n",
            "auth\trequisite\t$cdc_pam_lib deny\n"
            );

        my $rc = process_pam_d_file($cfg_file, \@lines, $libs_excluded, 
                    \@lines, $cdc_pam_lib, $cdc_long_comment, undef);
        if (!$rc)
        {
            # meet error
            $ret = $rc;
        }
    }

    # update "passwd" file
    # add keyword try_first_pass
    $cfg_file = "$cfg_dir/passwd";
    if (-e $cfg_file && !(grep /^$cfg_file$/, @$pam_d_ignored))
    {
        my @lines = @$newLines;
        @lines = grep {$_ !~ /\bauth\b/} @lines;

        unshift @lines, (
            "auth\tsufficient\t$cdc_pam_lib try_first_pass\n",
            "auth\trequisite\t$cdc_pam_lib deny\n"
            );

        my $rc = process_pam_d_file($cfg_file, \@lines, $libs_excluded, 
                    \@lines, $cdc_pam_lib, $cdc_long_comment, undef);
        if (!$rc)
        {
            # meet error
            $ret = $rc;
        }
    }

    # process all other files
    if (scalar(@cfg_files))
    {
        my $rc = process_pam_d_files(\@cfg_files, $cdc_pam_lib,
                    $libs_excluded, $cfg_dir, $cdc_long_comment, 
                    $newLines, $newLines2, undef);
        # got error or did not find any pam.d file ?
        if (!$rc)
        {
            # meet error
            $ret = $rc;
        }
    }

    return $ret;
}

my $ret = process_pam_d_general($cdc_pam_lib, $libs_excluded, "$cfg_dir/pam.d", 
            $cdc_long_comment, $newLines, $newLines2, $pam_d_ignored);

exit ($ret ? 0 : 1);

