#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2018 Centrify Corporation. All rights reserved.
#
# This script is used to modify nsswitch.conf
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error (general error)
#       2 if target file is empty
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts/da";
use CentrifyDA::Include;
use Getopt::Long qw(GetOptions);

use autoedit_common_utils qw(loadSetting getValueList flushNameService);
use autoedit_file_lock;
use CentrifyDA::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify nsswitch.conf
    usage:
    --lib_name          the name of centrifydc or centrifyda lib
    --cfg_file          the full path of nsswitch.conf
    --nss_xmaps         the maps to be updated
    --xxx_reaction      the reaction of map xxx
    --is_nss_active     check if DirectAudit NSS module is active 
                        this option has to be used together with "--lib_name centrifyda"
    --is_nss_normal     check if DirectAudit NSS module is as our expectation when it is enabled
                        this option has to be used together with "--lib_name centrifyda"
    --not_joined        indicates the machine is not joined
                        this option has to be used together with "--lib_name centrifyda"
    --do_load_cdc_conf  set it if want to load the above settings
                        from centrifydc.conf
    --do_revert         set it if want to revert our changes
END
    exit 1;
}

my @PRIORITY = ("centrifyda", "centrifydc");
use constant DEFAULT_CFG_FILE => "/etc/nsswitch.conf";
use constant DEFAULT_NSS_XMAPS => [];
use constant DEFAULT_CDC_TAG => "centrifydc";

my $cfgName = DEFAULT_CFG_FILE;
my $cfgHandler;
my $newName;
my $newHandler;

my $lib_name = DEFAULT_CDC_TAG;
my $nss_xmaps;
my $is_nss_active;
my $is_nss_normal;
my $not_joined = 0;
my $do_load_cdc_conf;
my $do_revert;

my @nssKeys = ( "passwd", "group", "shadow",
                "netgroup", "protocols", "rpc", "hosts", "initgroups", 
                "ethers", "ipnodes", "aliases", "services", 
                "prof_attr", "auth_attr", "bootparams", "netmasks", 
                "printers", "project", "networks");

my %nss_reactions = ();
my %nss_reaction_opts = ();
foreach (@nssKeys)
{
    $nss_reactions{$_} = undef;
    $nss_reaction_opts{"$_"."_reaction=s"} = \$nss_reactions{$_};
}

#
#
# Check if "centrifyda" is found in "passwd" key in nsswitch.conf
#
# $_[0]: the array ref contains all lines of nsswitch.conf
#
# Return value:
#       0 - Means that it is found
#       1 - Means that it is not found
#
sub isNssHooked
{
    my ($cfgLines) = @_;
    
    foreach my $line (@$cfgLines)
    {
        # it is not a comment
        if ($line =~ /^\s*(\w+)\s*:/)
        {
            my $key = $1;
            
            if (($key eq "passwd") && ($line =~ /\bcentrifyda\b/))
            {
                # Means that DirectAudit NSS module is active
                return 0;
            }
        }    
    }
    
    1;
}

#
#
# Check if DA nss status is as our expectation in nsswitch.conf when DA nss is enabled.
#
# for example:
# passwd: centrifydc centrifyda
# we think this is abnormal
#
# $_[0]: the array ref contains all lines of nsswitch.conf
#
# Return value:
#       0 - Means normal
#       1 - Means abnormal
#
sub isNssNormal
{
    my ($cfgLines) = @_;
    
    foreach my $line (@$cfgLines)
    {
        # it is not a comment
        if ($line =~ /^\s*(\w+)\s*:/)
        {
            my $key = $1;
            if ($key eq "passwd")
            {
                if(!$not_joined)
                {
                    # Means that CDC NSS module is enabled
                    if($line =~ /^\s*\w+\s*:\s*centrifyda\s*centrifydc\s*/)
                    {
                        return 0;
                    }
                }
                else
                {
                    # Means that CDC NSS module is not enabled
                    if(($line =~ /^\s*\w+\s*:\s*centrifyda\s*/) && ($line !~ /\bcentrifydc\b/))
                    {
                        return 0;
                    }
                }
            }           
        }       
    }
    
    1;
}

GetOptions( "lib_name=s" => \$lib_name,
            "cfg_file=s" => \$cfgName,
            "nss_xmaps=s" => \$nss_xmaps,
            "is_nss_active" => \$is_nss_active,
            "is_nss_normal" => \$is_nss_normal,
            "not_joined" => \$not_joined,
            "do_load_cdc_conf" => \$do_load_cdc_conf,
            "do_revert" => \$do_revert,
            %nss_reaction_opts
            ) or usage;

my %keysToCheck = ();

# the priority for the lib_name we support
# it may has multiple different lib_names in one line
# we need to sort them by priority
my %priority = ();
if (grep /^$lib_name$/, @PRIORITY)
{
    my $count = 0;
    foreach my $item (@PRIORITY)
    {
        $priority{$item} = $count;
        $count++;
    }
}
else
{
    # it is an unknown $lib_name
    # place it at top
    $priority{$lib_name} = 0;
}

# need not to check all options when do revert
if (!$do_revert)
{
    if (defined $nss_xmaps)
    {
        $nss_xmaps = [split /[,\s]+/, $nss_xmaps];
    }
    else
    {
        if ($do_load_cdc_conf)
        {
            $nss_xmaps = getValueList("adclient.autoedit.nss.xmaps");
        }

        if (!defined $nss_xmaps)
        {
            $nss_xmaps = DEFAULT_NSS_XMAPS;
        }
    }

    # get reaction for nss maps
    foreach my $key (@nssKeys)
    {
        if (!defined $nss_reactions{$key})
        {
            if ($do_load_cdc_conf)
            {
                # Get the user specified nss db reaction using key:
                # adclient.autoedit.nss.<name>.reaction: <value>
                # Ensure the reaction is enclosed with [ ]. 
                my $array_ref = getValueList("adclient.autoedit.nss.$key.reaction");
                if ($array_ref)
                {
                    my $reaction = $array_ref->[0];
                    if ($reaction && $reaction =~ /^\[[^\[\]]+\]$/)
                    {
                        $nss_reactions{$key} = $reaction;
                    }
                }
            }
        }
    }

    # the default reaction for "netgroup"
    if (!defined $nss_reactions{"netgroup"})
    {
        $nss_reactions{"netgroup"} = "[NOTFOUND=return]";
    }

    # determine which maps need to update
    if ($lib_name eq "centrifydc")
    {
        %keysToCheck = ('passwd' => 1,
                        'group' => 1,
                        'shadow' => 1,
                        # for IRIX, if "shadow" is not found, 
                        # try "shadow(mode=0700):"
                        'shadow(mode=0700)' => 1,
                        # Fedora has ill-documented initgroups after fc15.
                        'initgroups' => 1,);
    }
    elsif ($lib_name eq "centrifyda")
    {
        %keysToCheck = ('passwd' => 1);
    }

    if (grep /^hosts$/, @$nss_xmaps)
    {
        $keysToCheck{"hosts"} = 1;
    }

    foreach my $key (@$nss_xmaps)
    {
        if (grep /^$key$/, @nssKeys)
        {
            $keysToCheck{$key} = 1;
        }
    }
}

my $flock = autoedit_file_lock::new("$cfgName");
if (!defined $flock)
{
    DEBUG_OUT("cannot lock file $cfgName");
    exit 1;
}

my $realName = readlink($cfgName);
my $slink = undef;
if (defined $realName)
{
    my $realName = TraverseSymLink($cfgName);
    if (!defined $realName)
    {
        ERROR_OUT("meet error when check symlink $cfgName");
        exit 1;
    }
    $slink = $cfgName;
    $cfgName = $realName;
}

my $cfgLines = ReadFile("$cfgName");
if (!defined $cfgLines)
{
    exit 1;
}

my $cfgfile_empty = IsFileEmpty($cfgName);
if ($cfgfile_empty)
{
    exit 2;
}

my @cfgLines = split "\n", $cfgLines;

# check if DirectAudit NSS module is active
if ($is_nss_active)
{
    if ($lib_name eq "centrifyda")
    {
        my $ret = isNssHooked(\@cfgLines);
        
        exit $ret;
    }
    else
    {
        usage;
    }
}

# check if DirectAudit NSS module is normal
if ($is_nss_normal)
{
    if ($lib_name eq "centrifyda")
    {
        my $ret = isNssNormal(\@cfgLines);
        
        exit $ret;
    }
    else
    {
        usage;
    }
}

# update cfg file
my @newLines = ();
foreach my $line (@cfgLines)
{
    # it is not a comment
    if ($line =~ /^\s*(\w+)\s*:/)
    {
        my $key = $1;

        if ($do_revert)
        {
            # this line is modified by us ?
            if ($line =~ /\b$lib_name\b/)
            {
                # remove $lib_name and its reaction (if have)
                $line =~ s/\s*$lib_name(\s*\[.*\])?\s*/ /g;
            }
        }
        elsif (!exists $keysToCheck{$key})
        {
            # not all maps need to be updated
            # Only the maps defined by "--nss_xmaps" are wanted
            # continue to next line
        }
        elsif ($line =~ /\b$lib_name\b/)
        {
            # have been updated ?
            # continue to next line
            delete $keysToCheck{$key};
        }
        else
        {
            # determine $reaction
            my $reaction = defined $nss_reactions{$key} ? " $nss_reactions{$key}" : "";
            # prepare our code to insert
            my @ours = ("") x (scalar (keys %priority));
            foreach my $item (keys %priority)
            {
                if ($line =~ s/\s*($item(\s*\[.*\])?)\s*/ /)
                {
                    $ours[$priority{$item}] = " $1";
                }
            }
            $ours[$priority{$lib_name}] = " $lib_name$reaction";
            my $ours = join "", @ours;

            # insert our code
            $line =~ s/^(\s*\w+\s*:)\s*(.*)/$1$ours $2/;
            delete $keysToCheck{$key};
        }
    }
    push @newLines, "$line\n";
}

if (!$do_revert)
{
    if (defined $keysToCheck{"passwd"})
    {
        push @newLines, "passwd: centrifydc files\n";
    }

    if (defined $keysToCheck{"group"})
    {
        push @newLines, "group: centrifydc files\n";
    }
}

my $newLines = join "", @newLines;

if ($newLines eq $cfgLines)
{
    DEBUG_OUT("no changes in $cfgName, no-op");
    exit 0;
}

my $backup = "$cfgName.pre_cdc";
if ($lib_name eq "centrifyda")
{
    $backup = "$cfgName.pre_cda";
}

if (!$do_revert)
{
    # don't backup empty config file
    if (!$cfgfile_empty)
    {
        if (!CopyFile($cfgName, $backup))
        {
            ERROR_OUT("Cannot backup $cfgName");
            exit 1;
        }
    }
    else
    {
        ERROR_OUT("Cannot backup $cfgName because it is empty");
        exit 1;
    }
}
else
{
    # install.sh and rpm/pkg scripts may override config files with
    # the backup files if found them exist.
    # To avoid affecting these code logic, we'd better delete
    # backup file after "do_revert"
    if (-e $backup && !$cfgfile_empty)
    {
        unlink $backup;
    }
}

# Do not make it worse writing to empty config file
# (empty config file already caused "internal injuries")
if (!$cfgfile_empty)
{
    if (!WriteFile($cfgName, $newLines))
    {
        ERROR_OUT("Cannot write $cfgName");
        exit 1;
    }
}
else
{
    ERROR_OUT("Cannot write $cfgName because it is empty");
    exit 1;
}

# make sure nsswitch.conf file permission is 644
# for more info, please see bug33864
if (! chmod(0644, $cfgName))
{
    WARN_OUT("Cannot set mode for $cfgName.");
    exit 1;
}

#
# flush name cache service
flushNameService();

exit 0;

