#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2018 Centrify Corporation. All rights reserved.
#
# This script is used to restore pam files
#
##############################################################################


use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts";
use CentrifyDC::Include;
use File::Copy qw(copy);

use autoedit_configure qw($cfg_dir $cdc_pam_lib
                          $pam_d_ignored);
use autoedit_constants qw($CDC_Modified_Comment);
use autoedit_file_lock;
use CentrifyDC::GP::General qw(:debug :system);

# this function is used to restore out changes to pam file
sub restore_pam_file
{
    my ($cfgName, $cdc_long_comment, $cdc_pam_lib) = @_;
    my $cfgHandler;
    my $tmpName;
    my $tmpHandler;

    my $flock = autoedit_file_lock::new("$cfgName");
    if (!defined $flock)
    {
        ERROR_OUT("cannot lock file $cfgName");
        return undef;
    }

    $cfgName = TraverseSymLink($cfgName);
    if (!defined $cfgName)
    {
        ERROR_OUT("meet error when check symlink $cfgName");
        return undef;
    }

    # to make logic simple, read all lines here
    my $cfgLines = ReadFile("$cfgName");
    if (!defined $cfgLines)
    {
        ERROR_OUT("cannot read $cfgName");
        return undef;
    }
    my $oldLines = $cfgLines;
    my @cfgLines = split "\n", $cfgLines;
    @cfgLines = grep {$_ .= "\n"} @cfgLines;

    # remove all lines containing cdc_pam_lib
    my $cdc_pam_key = $cdc_pam_lib;
    # if it is a HPUX machine and not HP_PARISC
    # $cdc_pam_lib = "/usr/lib/security/\$ISA/libpam_centrifydc.so.1";
    # we don't need the FULL path, just want the file name without extension.
    $cdc_pam_key =~ s/.*\///;
    $cdc_pam_key =~ s/\..*//;

    # match if any word in the line starts with the key
    @cfgLines = grep {$_ !~ /^[^#]*\b$cdc_pam_key/} @cfgLines;

    # remove dzdo*
    @cfgLines = grep {$_ !~ /^[^#]*\bdzdo/} @cfgLines;

    # remove sshd-sso-mfa
    @cfgLines = grep {$_ !~ /^[^#]*\bsshd-sso-mfa/} @cfgLines;

    # remove comment
    # @cfgLines = grep {$_ ne $cdc_long_comment ? $_ : ""} @cfgLines;
    @cfgLines = grep {$_ !~ /$cdc_long_comment/} @cfgLines;
    
    # remove other keyword added by us, 
    # e.g. try_first_pass use_first_pass
    # We do that only for HPUX libs such as pam_unix.so, pam_unix.so.1.
    # That is because we remove other lines we added 
    # depending on keywords "libpam_centrifydc".
    # But due to our workaround for "requisite" on HPUX,
    # we have to revert these *_first_pass carefully here.
    my $pam_unix = "lib(pam_unix\\.so\\.1|pam_unix\\.1|pam_unix\\.so)";
    grep s/(^[^#]*\spassword\s.*\b$pam_unix\b.*)\stry_first_pass/$1/, @cfgLines;
    grep s/(^[^#]*\sauth\s.*\b$pam_unix\b.*)\suse_first_pass/$1/, @cfgLines;

    $cfgLines = join "", @cfgLines;

    if ($oldLines ne $cfgLines)
    {
        if (!WriteFile($cfgName, $cfgLines))
        {
            ERROR_OUT("Cannot write $cfgName");
            return undef;
        }
    }

    my $backup = "$cfgName.pre_cdc";
    if (-e $backup)
    {
        unlink("$backup");
    }

    return 1;
}

sub restore_pam_dir
{
    my ($cfg_dir, $cdc_long_comment, 
        $cdc_pam_lib, $pam_d_ignored) = @_;

    my @cfg_files = <$cfg_dir/*>;

    # Skip pre_cdc file and our internal PAM config file
    # e.g. centrifydc centrifypam on Darwin
    # Skip $pam_d_ignored
    {
        my @files = ();
        foreach my $file (@cfg_files)
        {
            if ($file =~ /(centrify|pre_cdc)/s)
            {
                next;
            }
            # Ignore the dzdo* files
            if ($file =~ /\/dzdo/)
            {
                next;
            }
            if (grep /^$file$/, @$pam_d_ignored)
            {
                next;
            }
            push @files, $file;
        }
        @cfg_files = @files;
    }

    foreach my $file (@cfg_files)
    {
        restore_pam_file($file, $cdc_long_comment, 
                         $cdc_pam_lib);
    }
}

my $cdc_long_comment = $CDC_Modified_Comment;
my $cfg_file = "$cfg_dir/pam.conf";
if (-e $cfg_file)
{
    restore_pam_file($cfg_file, $cdc_long_comment,
                    $cdc_pam_lib);
}

if (-e "$cfg_dir/pam.d")
{
    restore_pam_dir("$cfg_dir/pam.d", $cdc_long_comment, 
                    $cdc_pam_lib);
}

foreach ("system-auth", "password-auth")
{
    my $cfgName = "$cfg_dir/pam.d/$_";

    if (-l $cfgName)
    {
        my $flock = autoedit_file_lock::new("$cfgName");
        if (!defined $flock)
        {
            DEBUG_OUT("cannot lock file $cfgName");
            exit 1;
        }

        my $realName = TraverseSymLink($cfgName);
        if ($realName =~ /\.cdc$/)
        {
            $realName = $`;
            rename "$realName.cdc", "$realName";
            unlink "$cfgName";
            symlink "$realName", "$cfgName";
            my $backupName = "$realName.pre_cdc";
            if (-e $backupName)
            {
                unlink($backupName);
            }
        }
    }
}

exit 0;
